#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
General purpose separator block for IDAES models
"""

from enum import Enum
from functools import partial
from pandas import DataFrame

from pyomo.environ import (
    Block,
    check_optimal_termination,
    Constraint,
    Param,
    Reals,
    Reference,
    Set,
    Var,
    value,
)
from pyomo.network import Port
from pyomo.common.config import ConfigBlock, ConfigValue, In, ListOf, Bool

from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
    MaterialBalanceType,
    MomentumBalanceType,
    MaterialFlowBasis,
    VarLikeExpression,
)
from idaes.core.util.config import (
    is_physical_parameter_block,
    is_state_block,
)
from idaes.core.util.exceptions import (
    BurntToast,
    ConfigurationError,
    PropertyNotSupportedError,
    InitializationError,
)
from idaes.core.solvers import get_solver
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale
from idaes.core.util.units_of_measurement import report_quantity
from idaes.core.initialization import ModularInitializerBase

__author__ = "Andrew Lee"


# Set up logger
_log = idaeslog.getLogger(__name__)


# Enumerate options for balances
class SplittingType(Enum):
    """
    Enum of supported material split types.
    """

    totalFlow = 1
    phaseFlow = 2
    componentFlow = 3
    phaseComponentFlow = 4


class EnergySplittingType(Enum):
    """
    Enum of support energy split types.
    """

    none = 0
    equal_temperature = 1
    equal_molar_enthalpy = 2
    enthalpy_split = 3


class SeparatorInitializer(ModularInitializerBase):
    """
    Initializer for Separator blocks.

    """

    def initialization_routine(
        self,
        model: Block,
    ):
        """
        Initialization routine for Separator Blocks.

        This routine starts by initializing the feed stream followed by solving
        for all split fractions. Next, states for each outlet are estimated using the
        following rules:

        1. Intensive states remain unchanged
        2. Extensive states are multiplied by split fractions if index matches, or
            average of split fractions for outlet otherwise

        Each outlet state block is then initialized, and finally the full model is
        solved.

        Args:
            model: model to be initialized

        Returns:
            None

        """
        init_log = idaeslog.getInitLogger(
            model.name, self.get_output_level(), tag="unit"
        )
        solve_log = idaeslog.getSolveLogger(
            model.name, self.get_output_level(), tag="unit"
        )

        # Create solver
        solver = self._get_solver()

        # Initialize mixed state block
        if model.config.mixed_state_block is not None:
            mblock = model.config.mixed_state_block
        else:
            mblock = model.mixed_state
        self.get_submodel_initializer(mblock).initialize(mblock)

        # Solve for split fractions only
        component_status = {}
        for c in model.component_objects((Block, Constraint)):
            for i in c:
                if not c[i].local_name == "sum_split_frac":
                    # Record current status of components to restore later
                    component_status[c[i]] = c[i].active
                    c[i].deactivate()

        res = None
        if degrees_of_freedom(model) != 0:
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = solver.solve(model, tee=slc.tee)
                init_log.info(
                    "Initialization Step 1 Complete: {}".format(idaeslog.condition(res))
                )

        for c, s in component_status.items():
            if s:
                c.activate()

        if model.config.ideal_separation:
            # If using ideal splitting, initialization should be complete
            return res

        # Initialize outlet StateBlocks
        outlet_list = model.create_outlet_list()

        # Premises for initializing outlet states:
        # 1. Intensive states remain unchanged - this is either a valid premise
        # or the actual state is impossible to calculate without solving the
        # full separator model.
        # 2. Extensive states use split fractions if index matches, or
        # average of split fractions for outlet otherwise
        for o in outlet_list:
            # Get corresponding outlet StateBlock
            o_block = getattr(model, o + "_state")

            # Create dict to store fixed status of state variables
            for t in model.flowsheet().time:
                # Calculate values for state variables
                s_vars = o_block[t].define_state_vars()
                for var_name_port, var_obj in s_vars.items():
                    for k in var_obj:
                        # If fixed, use current value
                        # otherwise calculate guess from mixed state and fix
                        if not var_obj[k].fixed:
                            m_var = getattr(mblock[t], var_obj.local_name)
                            if "flow" in var_name_port:
                                # If a "flow" variable, is extensive
                                # Apply split fraction
                                if model.config.split_basis == SplittingType.totalFlow:
                                    # All flows split by outlet
                                    var_obj[k].set_value(
                                        value(m_var[k] * model.split_fraction[(t, o)])
                                    )
                                elif "_phase_comp" in var_name_port:
                                    # Need to match indices, but use split frac
                                    if (
                                        model.config.split_basis
                                        == SplittingType.phaseComponentFlow
                                    ):
                                        var_obj[k].set_value(
                                            value(
                                                m_var[k]
                                                * model.split_fraction[(t, o) + (k,)]
                                            )
                                        )
                                    elif (
                                        model.config.split_basis
                                        == SplittingType.phaseFlow
                                    ):
                                        var_obj[k].set_value(
                                            value(
                                                m_var[k]
                                                * model.split_fraction[(t, o) + (k[0],)]
                                            )
                                        )
                                    elif (
                                        model.config.split_basis
                                        == SplittingType.componentFlow
                                    ):
                                        var_obj[k].set_value(
                                            value(
                                                m_var[k]
                                                * model.split_fraction[(t, o) + (k[1],)]
                                            )
                                        )
                                    else:
                                        raise BurntToast(
                                            "{} encountered unrecognised "
                                            "SplittingType. This should not "
                                            "occur - please send this bug to "
                                            "the IDAES developers.".format(model.name)
                                        )
                                elif "_phase" in var_name_port:
                                    if (
                                        model.config.split_basis
                                        == SplittingType.phaseComponentFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                model.split_fraction[t, o, k, j]
                                                for j in mblock.component_list
                                            )
                                            / len(mblock.component_list)
                                        )
                                        var_obj[k].set_value(
                                            value(m_var[k] * avg_split)
                                        )
                                    elif (
                                        model.config.split_basis
                                        == SplittingType.phaseFlow
                                    ):
                                        var_obj[k].set_value(
                                            value(
                                                m_var[k]
                                                * model.split_fraction[(t, o) + (k,)]
                                            )
                                        )
                                    elif (
                                        model.config.split_basis
                                        == SplittingType.componentFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                model.split_fraction[t, o, j]
                                                for j in mblock.component_list
                                            )
                                            / len(mblock.component_list)
                                        )
                                        var_obj[k].set_value(
                                            value(m_var[k] * avg_split)
                                        )
                                    else:
                                        raise BurntToast(
                                            "{} encountered unrecognised "
                                            "SplittingType. This should not "
                                            "occur - please send this bug to "
                                            "the IDAES developers.".format(model.name)
                                        )
                                elif "_comp" in var_name_port:
                                    if (
                                        model.config.split_basis
                                        == SplittingType.phaseComponentFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                model.split_fraction[t, o, p, k]
                                                for p in mblock.phase_list
                                            )
                                            / len(mblock.phase_list)
                                        )
                                        var_obj[k].set_value(
                                            value(m_var[k] * avg_split)
                                        )
                                    elif (
                                        model.config.split_basis
                                        == SplittingType.phaseFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                model.split_fraction[t, o, p]
                                                for p in mblock.phase_list
                                            )
                                            / len(mblock.phase_list)
                                        )
                                        var_obj[k].set_value(
                                            value(m_var[k] * avg_split)
                                        )
                                    elif (
                                        model.config.split_basis
                                        == SplittingType.componentFlow
                                    ):
                                        var_obj[k].set_value(
                                            value(
                                                m_var[k]
                                                * model.split_fraction[(t, o) + (k,)]
                                            )
                                        )
                                    else:
                                        raise BurntToast(
                                            "{} encountered unrecognised "
                                            "SplittingType. This should not "
                                            "occur - please send this bug to "
                                            "the IDAES developers.".format(model.name)
                                        )
                                else:
                                    # Assume unindexed extensive state
                                    # Need average split
                                    if (
                                        model.config.split_basis
                                        == SplittingType.phaseComponentFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                model.split_fraction[t, o, p, j]
                                                for (p, j) in mblock.phase_component_set
                                            )
                                            / len(mblock.phase_component_set)
                                        )
                                    elif (
                                        model.config.split_basis
                                        == SplittingType.phaseFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                model.split_fraction[t, o, p]
                                                for p in mblock.phase_list
                                            )
                                            / len(mblock.phase_list)
                                        )
                                    elif (
                                        model.config.split_basis
                                        == SplittingType.componentFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                model.split_fraction[t, o, j]
                                                for j in mblock.component_list
                                            )
                                            / len(mblock.component_list)
                                        )
                                    else:
                                        raise BurntToast(
                                            "{} encountered unrecognised "
                                            "SplittingType. This should not "
                                            "occur - please send this bug to "
                                            "the IDAES developers.".format(model.name)
                                        )
                                    var_obj[k].set_value(value(m_var[k] * avg_split))
                            else:
                                # Otherwise intensive, equate to mixed stream
                                var_obj[k].set_value(m_var[k].value)

            # Call initialization routine for outlet StateBlock
            self.get_submodel_initializer(o_block).initialize(o_block)

        if model.config.mixed_state_block is None:
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = solver.solve(model, tee=slc.tee)

            if not check_optimal_termination(res):
                raise InitializationError(
                    f"{model.name} failed to initialize successfully. Please "
                    f"check the output logs for more information."
                )

            init_log.info(
                "Initialization Step 2 Complete: {}".format(idaeslog.condition(res))
            )
        else:
            init_log.info("Initialization Complete.")

        return res


@declare_process_block_class("Separator")
class SeparatorData(UnitModelBlockData):
    """
    This is a general purpose model for a Separator block with the IDAES
    modeling framework. This block can be used either as a stand-alone
    Separator unit operation, or as a sub-model within another unit operation.

    This model creates a number of StateBlocks to represent the outgoing
    streams, then writes a set of phase-component material balances, an
    overall enthalpy balance (2 options), and a momentum balance (2 options)
    linked to a mixed-state StateBlock. The mixed-state StateBlock can either
    be specified by the user (allowing use as a sub-model), or created by the
    Separator.

    When being used as a sub-model, Separator should only be used when a
    set of new StateBlocks are required for the streams to be separated. It
    should not be used to separate streams to go to multiple ControlVolumes in a
    single unit model - in these cases the unit model developer should write
    their own splitting equations.
    """

    default_initializer = SeparatorInitializer

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
**default** = False. Product blocks are always steady-state.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Product blocks do not contain holdup, thus this must be
False.""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for mixer",
            doc="""Property parameter object used to define property
calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property
block(s) and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "outlet_list",
        ConfigValue(
            domain=ListOf(str),
            description="List of outlet names",
            doc="""A list containing names of outlets,
**default** - None.
**Valid values:** {
**None** - use num_outlets argument,
**list** - a list of names to use for outlets.}""",
        ),
    )
    CONFIG.declare(
        "num_outlets",
        ConfigValue(
            domain=int,
            description="Number of outlets to unit",
            doc="""Argument indicating number (int) of outlets to construct,
not used if outlet_list arg is provided,
**default** - None.
**Valid values:** {
**None** - use outlet_list arg instead, or default to 2 if neither argument
provided,
**int** - number of outlets to create (will be named with sequential integers
from 1 to num_outlets).}""",
        ),
    )
    CONFIG.declare(
        "split_basis",
        ConfigValue(
            default=SplittingType.totalFlow,
            domain=SplittingType,
            description="Basis for splitting stream",
            doc="""Argument indicating basis to use for splitting mixed stream,
**default** - SplittingType.totalFlow.
**Valid values:** {
**SplittingType.totalFlow** - split based on total flow (split
fraction indexed only by time and outlet),
**SplittingType.phaseFlow** - split based on phase flows (split fraction
indexed by time, outlet and phase),
**SplittingType.componentFlow** - split based on component flows (split
fraction indexed by time, outlet and components),
**SplittingType.phaseComponentFlow** - split based on phase-component flows (
split fraction indexed by both time, outlet, phase and components).}""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.useDefault.
**Valid values:** {
**MaterialBalanceType.useDefault - refer to property package for default
balance type
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
    **default** - MomentumBalanceType.pressureTotal.
    **Valid values:** {
    **MomentumBalanceType.none** - exclude momentum balances,
    **MomentumBalanceType.pressureTotal** - pressure in all outlets is equal,
    **MomentumBalanceType.pressurePhase** - not yet supported,
    **MomentumBalanceType.momentumTotal** - not yet supported,
    **MomentumBalanceType.momentumPhase** - not yet supported.}""",
        ),
    )
    CONFIG.declare(
        "has_phase_equilibrium",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Calculate phase equilibrium in mixed stream",
            doc="""Argument indicating whether phase equilibrium should be
calculated for the resulting mixed stream,
**default** - False.
**Valid values:** {
**True** - calculate phase equilibrium in mixed stream,
**False** - do not calculate equilibrium in mixed stream.}""",
        ),
    )
    CONFIG.declare(
        "energy_split_basis",
        ConfigValue(
            default=EnergySplittingType.equal_temperature,
            domain=EnergySplittingType,
            description="Type of constraint to write for energy splitting",
            doc="""Argument indicating basis to use for splitting energy this
is not used for when ideal_separation == True.
**default** - EnergySplittingType.equal_temperature.
**Valid values:** {
**EnergySplittingType.none** - no energy balance constraints,
**EnergySplittingType.equal_temperature** - outlet temperatures equal inlet,
**EnergySplittingType.equal_molar_enthalpy** - outlet molar enthalpies equal
inlet,
**EnergySplittingType.enthalpy_split** - apply split fractions to enthalpy
flows. Does not work with component or phase-component splitting.}""",
        ),
    )
    CONFIG.declare(
        "ideal_separation",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Ideal splitting flag",
            doc="""Argument indicating whether ideal splitting should be used.
Ideal splitting assumes perfect spearation of material, and attempts to
avoid duplication of StateBlocks by directly partitioning outlet flows to
ports,
**default** - False.
**Valid values:** {
**True** - use ideal splitting methods. Cannot be combined with
has_phase_equilibrium = True,
**False** - use explicit splitting equations with split fractions.}""",
        ),
    )
    CONFIG.declare(
        "ideal_split_map",
        ConfigValue(
            domain=dict,
            description="Ideal splitting partitioning map",
            doc="""Dictionary containing information on how extensive variables
should be partitioned when using ideal splitting (ideal_separation = True).
**default** - None.
**Valid values:** {
**dict** with keys of indexing set members and values indicating which outlet
this combination of keys should be partitioned to.
E.g. {("Vap", "H2"): "outlet_1"}}""",
        ),
    )
    CONFIG.declare(
        "mixed_state_block",
        ConfigValue(
            domain=is_state_block,
            description="Existing StateBlock to use as mixed stream",
            doc="""An existing state block to use as the source stream from the
Separator block,
**default** - None.
**Valid values:** {
**None** - create a new StateBlock for the mixed stream,
**StateBlock** - a StateBock to use as the source for the mixed stream.}""",
        ),
    )
    CONFIG.declare(
        "construct_ports",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Construct inlet and outlet Port objects",
            doc="""Argument indicating whether model should construct Port
objects linked the mixed state and all outlet states,
**default** - True.
**Valid values:** {
**True** - construct Ports for all states,
**False** - do not construct Ports.""",
        ),
    )

    def build(self):
        """
        General build method for SeparatorData. This method calls a number
        of sub-methods which automate the construction of expected attributes
        of unit models.

        Inheriting models should call `super().build`.

        Args:
            None

        Returns:
            None
        """
        # Call super.build()
        super(SeparatorData, self).build()

        self._validate_config_arguments()

        # Call setup methods from ControlVolumeBlockData
        self._get_property_package()
        self._get_indexing_sets()

        # Create list of inlet names
        outlet_list = self.create_outlet_list()

        if self.config.mixed_state_block is None:
            mixed_block = self.add_mixed_state_block()
        else:
            mixed_block = self.get_mixed_state_block()

        # Add inlet port
        self.add_inlet_port_objects(mixed_block)

        # Construct splitter based on ideal_separation argument
        if self.config.ideal_separation:
            # Use ideal partitioning method
            self.partition_outlet_flows(mixed_block, outlet_list)
        else:
            # Otherwise, Build StateBlocks for outlet
            outlet_blocks = self.add_outlet_state_blocks(outlet_list)

            # Add split fractions
            self.add_split_fractions(outlet_list, mixed_block)

            # Construct splitting equations
            self.add_material_splitting_constraints(mixed_block)
            self.add_energy_splitting_constraints(mixed_block)
            self.add_momentum_splitting_constraints(mixed_block)

            # Construct outlet port objects
            self.add_outlet_port_objects(outlet_list, outlet_blocks)

    def _validate_config_arguments(self):
        if self.config.has_phase_equilibrium and self.config.ideal_separation:
            raise ConfigurationError(
                """{} received arguments has_phase_equilibrium = True and
                    ideal_separation = True. These arguments are incompatible
                    with each other, and you should choose one or the other.""".format(
                    self.name
                )
            )

    def create_outlet_list(self):
        """
        Create list of outlet stream names based on config arguments.

        Returns:
            list of strings
        """
        if self.config.outlet_list is not None and self.config.num_outlets is not None:
            # If both arguments provided and not consistent, raise Exception
            if len(self.config.outlet_list) != self.config.num_outlets:
                raise ConfigurationError(
                    "{} Separator provided with both outlet_list and "
                    "num_outlets arguments, which were not consistent ("
                    "length of outlet_list was not equal to num_outlets). "
                    "Please check your arguments for consistency, and "
                    "note that it is only necessry to provide one of "
                    "these arguments.".format(self.name)
                )
        elif self.config.outlet_list is None and self.config.num_outlets is None:
            # If no arguments provided for outlets, default to num_outlets = 2
            self.config.num_outlets = 2

        # Create a list of names for outlet StateBlocks
        if self.config.outlet_list is not None:
            outlet_list = self.config.outlet_list
        else:
            outlet_list = [
                "outlet_" + str(n) for n in range(1, self.config.num_outlets + 1)
            ]

        return outlet_list

    def add_outlet_state_blocks(self, outlet_list):
        """
        Construct StateBlocks for all outlet streams.

        Args:
            list of strings to use as StateBlock names

        Returns:
            list of StateBlocks
        """
        # Setup StateBlock argument dict
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["defined_state"] = False

        # Create empty list to hold StateBlocks for return
        outlet_blocks = []

        # Create an instance of StateBlock for all outlets
        for o in outlet_list:
            o_obj = self.config.property_package.build_state_block(
                self.flowsheet().time, doc="Material properties at outlet", **tmp_dict
            )

            setattr(self, o + "_state", o_obj)

            outlet_blocks.append(getattr(self, o + "_state"))

        return outlet_blocks

    def add_mixed_state_block(self):
        """
        Constructs StateBlock to represent mixed stream.

        Returns:
            New StateBlock object
        """
        # Setup StateBlock argument dict
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["defined_state"] = True

        self.mixed_state = self.config.property_package.build_state_block(
            self.flowsheet().time, doc="Material properties of mixed stream", **tmp_dict
        )

        return self.mixed_state

    def get_mixed_state_block(self):
        """
        Validates StateBlock provided in user arguments for mixed stream.

        Returns:
            The user-provided StateBlock or an Exception
        """
        # Sanity check to make sure method is not called when arg missing
        if self.config.mixed_state_block is None:
            try:
                return self.mixed_state
            except AttributeError:
                raise BurntToast(
                    f"{self.name} get_mixed_state_block method called when the "
                    "mixed_state_block argument is None, and no mixed_state "
                    "block is contained in separator. This should not happen."
                )
        # Check that the user-provided StateBlock uses the same prop pack
        if (
            self.config.mixed_state_block[
                self.flowsheet().time.first()
            ].config.parameters
            != self.config.property_package
        ):
            raise ConfigurationError(
                "{} StateBlock provided in mixed_state_block argument "
                " does not come from the same property package as "
                "provided in the property_package argument. All "
                "StateBlocks within a Separator must use the same "
                "property package.".format(self.name)
            )

        return self.config.mixed_state_block

    def add_inlet_port_objects(self, mixed_block):
        """
        Adds inlet Port object if required.

        Args:
            a mixed state StateBlock object

        Returns:
            None
        """
        if self.config.construct_ports is True:
            self.add_port(name="inlet", block=mixed_block, doc="Inlet Port")

    def add_outlet_port_objects(self, outlet_list, outlet_blocks):
        """
        Adds outlet Port objects if required.

        Args:
            a list of outlet StateBlock objects

        Returns:
            None
        """
        if self.config.construct_ports is True:
            # Add ports
            for p in outlet_list:
                o_state = getattr(self, p + "_state")
                self.add_port(name=p, block=o_state, doc="Outlet Port")

    def add_split_fractions(self, outlet_list, mixed_block):
        """
        Creates outlet Port objects and tries to partition mixed stream flows
        between these

        Args:
            StateBlock representing the mixed flow to be split
            a list of names for outlets

        Returns:
            None
        """
        self.outlet_idx = Set(initialize=outlet_list)
        pc_set = mixed_block.phase_component_set

        if self.config.split_basis == SplittingType.totalFlow:
            sf_idx = [self.flowsheet().time, self.outlet_idx]
            sf_sum_idx = [self.flowsheet().time]
        elif self.config.split_basis == SplittingType.phaseFlow:
            sf_idx = [
                self.flowsheet().time,
                self.outlet_idx,
                mixed_block.phase_list,
            ]
            sf_sum_idx = [
                self.flowsheet().time,
                mixed_block.phase_list,
            ]
        elif self.config.split_basis == SplittingType.componentFlow:
            sf_idx = [
                self.flowsheet().time,
                self.outlet_idx,
                mixed_block.component_list,
            ]
            sf_sum_idx = [
                self.flowsheet().time,
                mixed_block.component_list,
            ]
        elif self.config.split_basis == SplittingType.phaseComponentFlow:
            sf_idx = [
                self.flowsheet().time,
                self.outlet_idx,
                pc_set,
            ]
            sf_sum_idx = [
                self.flowsheet().time,
                pc_set,
            ]
        else:
            raise BurntToast(
                "{} split_basis has unexpected value. This "
                "should not happen.".format(self.name)
            )

        # Create split fraction variable
        self.split_fraction = Var(*sf_idx, initialize=0.5, doc="Outlet split fractions")

        # Add constraint that split fractions sum to 1
        def sum_sf_rule(b, t, *args):
            return 1 == sum(b.split_fraction[t, o, args] for o in self.outlet_idx)

        self.sum_split_frac = Constraint(*sf_sum_idx, rule=sum_sf_rule)

    def add_material_splitting_constraints(self, mixed_block):
        """
        Creates constraints for splitting the material flows
        """
        pc_set = mixed_block.phase_component_set

        def sf(t, o, p, j):
            if self.config.split_basis == SplittingType.totalFlow:
                return self.split_fraction[t, o]
            elif self.config.split_basis == SplittingType.phaseFlow:
                return self.split_fraction[t, o, p]
            elif self.config.split_basis == SplittingType.componentFlow:
                return self.split_fraction[t, o, j]
            elif self.config.split_basis == SplittingType.phaseComponentFlow:
                return self.split_fraction[t, o, p, j]

        mb_type = self.config.material_balance_type
        if mb_type == MaterialBalanceType.useDefault:
            t_ref = self.flowsheet().time.first()
            mb_type = mixed_block[t_ref].default_material_balance_type()

        # Get units from property package
        units_meta = mixed_block.params.get_metadata()
        flow_basis = mixed_block[
            self.flowsheet().time.first()
        ].get_material_flow_basis()
        if flow_basis == MaterialFlowBasis.molar:
            flow_units = units_meta.get_derived_units("flow_mole")
        elif flow_basis == MaterialFlowBasis.mass:
            flow_units = units_meta.get_derived_units("flow_mass")
        else:
            # Let this pass for now with no units
            flow_units = None

        if mixed_block.include_inherent_reactions:
            # Add extents of reaction and stoichiometric constraints
            # TODO: It would be nice if there was a way to recognise that
            # a total flow split does not require calculation of equilibrium
            # (assuming T&P are constant) and turn of inherent reactions in
            # that case, but that feature does not exist in property packages
            self.inherent_reaction_extent = Var(
                self.flowsheet().time,
                self.outlet_idx,
                mixed_block.params.inherent_reaction_idx,
                domain=Reals,
                initialize=0.0,
                doc="Extent of inherent reactions in outlets",
                units=flow_units,
            )

            self.inherent_reaction_generation = Var(
                self.flowsheet().time,
                self.outlet_idx,
                pc_set,
                domain=Reals,
                initialize=0.0,
                doc="Generation due to inherent reactions in outlets",
                units=flow_units,
            )

            @self.Constraint(
                self.flowsheet().time,
                self.outlet_idx,
                pc_set,
            )
            def inherent_reaction_constraint(b, t, o, p, j):
                if (p, j) in pc_set:
                    return b.inherent_reaction_generation[t, o, p, j] == (
                        sum(
                            mixed_block[t].params.inherent_reaction_stoichiometry[
                                r, p, j
                            ]
                            * self.inherent_reaction_extent[t, o, r]
                            for r in mixed_block[t].params.inherent_reaction_idx
                        )
                    )
                return Constraint.Skip

        if mb_type == MaterialBalanceType.componentPhase:
            if self.config.has_phase_equilibrium is True:
                try:
                    self.phase_equilibrium_generation = Var(
                        self.flowsheet().time,
                        self.outlet_idx,
                        self.config.property_package.phase_equilibrium_idx,
                        domain=Reals,
                        doc="Amount of generation in unit by phase equilibria",
                        units=flow_units,
                    )
                except AttributeError:
                    raise PropertyNotSupportedError(
                        "{} Property package does not contain a list of phase "
                        "equilibrium reactions (phase_equilibrium_idx), thus "
                        "does not support phase equilibrium.".format(self.name)
                    )

            @self.Constraint(
                self.flowsheet().time,
                self.outlet_idx,
                pc_set,
                doc="Material splitting equations",
            )
            def material_splitting_eqn(b, t, o, p, j):
                o_block = getattr(self, o + "_state")
                lhs = sf(t, o, p, j) * mixed_block[t].get_material_flow_terms(p, j)

                rhs = o_block[t].get_material_flow_terms(p, j)

                if self.config.has_phase_equilibrium:
                    rhs += -sum(
                        b.phase_equilibrium_generation[t, o, r]
                        for r in b.config.property_package.phase_equilibrium_idx
                        if mixed_block[t].params.phase_equilibrium_list[r][0] == j
                        and mixed_block[t].params.phase_equilibrium_list[r][1][0] == p
                    ) + sum(
                        b.phase_equilibrium_generation[t, o, r]
                        for r in b.config.property_package.phase_equilibrium_idx
                        if mixed_block[t].params.phase_equilibrium_list[r][0] == j
                        and mixed_block[t].params.phase_equilibrium_list[r][1][1] == p
                    )

                if mixed_block.include_inherent_reactions:
                    rhs += b.inherent_reaction_generation[t, o, p, j]

                return lhs == rhs

        elif mb_type == MaterialBalanceType.componentTotal:

            @self.Constraint(
                self.flowsheet().time,
                self.outlet_idx,
                mixed_block.component_list,
                doc="Material splitting equations",
            )
            def material_splitting_eqn(b, t, o, j):
                o_block = getattr(self, o + "_state")

                lhs = sum(
                    sf(t, o, p, j) * mixed_block[t].get_material_flow_terms(p, j)
                    for p in mixed_block.phase_list
                    if (p, j) in pc_set
                )
                rhs = sum(
                    o_block[t].get_material_flow_terms(p, j)
                    for p in o_block.phase_list
                    if (p, j) in pc_set
                )

                if mixed_block.include_inherent_reactions:
                    rhs += sum(
                        b.inherent_reaction_generation[t, o, p, j]
                        for p in o_block.phase_list
                        if (p, j) in pc_set
                    )

                return lhs == rhs

        elif mb_type == MaterialBalanceType.total:

            @self.Constraint(
                self.flowsheet().time,
                self.outlet_idx,
                doc="Material splitting equations",
            )
            def material_splitting_eqn(b, t, o):
                o_block = getattr(self, o + "_state")

                lhs = sum(
                    sum(
                        sf(t, o, p, j) * mixed_block[t].get_material_flow_terms(p, j)
                        for j in mixed_block.component_list
                        if (p, j) in pc_set
                    )
                    for p in mixed_block.phase_list
                )
                rhs = sum(
                    sum(
                        o_block[t].get_material_flow_terms(p, j)
                        for j in mixed_block.component_list
                        if (p, j) in pc_set
                    )
                    for p in o_block.phase_list
                )

                if mixed_block.include_inherent_reactions:
                    rhs += sum(
                        sum(
                            b.inherent_reaction_generation[t, o, p, j]
                            for j in mixed_block.component_list
                            if (p, j) in pc_set
                        )
                        for p in o_block.phase_list
                    )

                return lhs == rhs

        elif mb_type == MaterialBalanceType.elementTotal:
            raise ConfigurationError(
                "{} Separators do not support elemental "
                "material balances.".format(self.name)
            )
        elif mb_type == MaterialBalanceType.none:
            pass
        else:
            raise BurntToast(
                "{} Separator received unrecognised value for "
                "material_balance_type. This should not happen, "
                "please report this bug to the IDAES developers.".format(self.name)
            )

    def add_energy_splitting_constraints(self, mixed_block):
        """
        Creates constraints for splitting the energy flows.
        """
        if self.config.energy_split_basis == EnergySplittingType.none:
            # Do nothing and return
            return
        elif self.config.energy_split_basis == EnergySplittingType.equal_temperature:

            @self.Constraint(
                self.flowsheet().time,
                self.outlet_idx,
                doc="Temperature equality constraint",
            )
            def temperature_equality_eqn(b, t, o):
                o_block = getattr(self, o + "_state")
                return mixed_block[t].temperature == o_block[t].temperature

        elif self.config.energy_split_basis == EnergySplittingType.equal_molar_enthalpy:

            @self.Constraint(
                self.flowsheet().time,
                self.outlet_idx,
                doc="Molar enthalpy equality constraint",
            )
            def molar_enthalpy_equality_eqn(b, t, o):
                o_block = getattr(self, o + "_state")
                return mixed_block[t].enth_mol == o_block[t].enth_mol

        elif self.config.energy_split_basis == EnergySplittingType.enthalpy_split:
            # Validate split fraction type
            if (
                self.config.split_basis == SplittingType.phaseComponentFlow
                or self.config.split_basis == SplittingType.componentFlow
            ):
                raise ConfigurationError(
                    "{} Cannot use energy_split_basis == enthalpy_split "
                    "with split_basis == component or phaseComponent.".format(self.name)
                )

            def sf(t, o, p):
                if self.config.split_basis == SplittingType.totalFlow:
                    return self.split_fraction[t, o]
                elif self.config.split_basis == SplittingType.phaseFlow:
                    return self.split_fraction[t, o, p]

            @self.Constraint(
                self.flowsheet().time,
                self.outlet_idx,
                doc="Molar enthalpy splitting constraint",
            )
            def molar_enthalpy_splitting_eqn(b, t, o):
                o_block = getattr(self, o + "_state")
                return sum(
                    mixed_block[t].get_enthalpy_flow_terms(p) * sf(t, o, p)
                    for p in mixed_block.phase_list
                ) == sum(
                    o_block[t].get_enthalpy_flow_terms(p) for p in o_block.phase_list
                )

        else:
            raise BurntToast(
                "{} received unrecognised value for "
                "energy_split_basis. This should never happen, so"
                " please contact the IDAES developers with this "
                "bug.".format(self.name)
            )

    def add_momentum_splitting_constraints(self, mixed_block):
        """
        Creates constraints for splitting the momentum flows - done by equating
        pressures in outlets.
        """

        if self.config.momentum_balance_type is MomentumBalanceType.none:
            pass
        elif self.config.momentum_balance_type is MomentumBalanceType.pressureTotal:

            @self.Constraint(
                self.flowsheet().time,
                self.outlet_idx,
                doc="Pressure equality constraint",
            )
            def pressure_equality_eqn(b, t, o):
                o_block = getattr(self, o + "_state")
                return mixed_block[t].pressure == o_block[t].pressure

        else:
            raise NotImplementedError(
                f"Separators do not yet support momentum balances of type "
                f"{self.config.momentum_balance_type}"
            )

    def partition_outlet_flows(self, mb, outlet_list):
        """
        Creates outlet Port objects and tries to partition mixed stream flows
        between these

        Args:
            StateBlock representing the mixed flow to be split
            a list of names for outlets

        Returns:
            None
        """
        # Check arguments
        if self.config.construct_ports is False:
            raise ConfigurationError(
                "{} cannot have and ideal separator "
                "(ideal_separation = True) with "
                "construct_ports = False.".format(self.name)
            )
        if self.config.split_basis == SplittingType.totalFlow:
            raise ConfigurationError(
                "{} cannot do an ideal separation based "
                "on total flow. Either use ideal_separation = False or a "
                "different separation basis.".format(self.name)
            )
        if self.config.ideal_split_map is None:
            raise ConfigurationError(
                "{} was not provided with an "
                "ideal_split_map argument which is "
                "necessary for doing an ideal_separation.".format(self.name)
            )

        # Validate split map
        split_map = self.config.ideal_split_map
        idx_list = []
        if self.config.split_basis == SplittingType.phaseFlow:
            for p in mb.phase_list:
                idx_list.append((p))

            if len(idx_list) != len(split_map):
                raise ConfigurationError(
                    "{} ideal_split_map does not match with "
                    "split_basis chosen. ideal_split_map must"
                    " have a key for each combination of indices.".format(self.name)
                )
            for k in idx_list:
                if k not in split_map:
                    raise ConfigurationError(
                        "{} ideal_split_map does not match with "
                        "split_basis chosen. ideal_split_map must"
                        " have a key for each combination of indices.".format(self.name)
                    )

        elif self.config.split_basis == SplittingType.componentFlow:
            for j in mb.component_list:
                idx_list.append((j))

            if len(idx_list) != len(split_map):
                raise ConfigurationError(
                    "{} ideal_split_map does not match with "
                    "split_basis chosen. ideal_split_map must"
                    " have a key for each component.".format(self.name)
                )
        elif self.config.split_basis == SplittingType.phaseComponentFlow:
            for p in mb.phase_list:
                for j in mb.component_list:
                    idx_list.append((p, j))

            if len(idx_list) != len(split_map):
                raise ConfigurationError(
                    "{} ideal_split_map does not match with "
                    "split_basis chosen. ideal_split_map must"
                    " have a key for each phase-component pair.".format(self.name)
                )

        # Check that no. outlets matches split_basis
        if len(outlet_list) != len(idx_list):
            raise ConfigurationError(
                "{} Cannot perform ideal separation. Must have one "
                "outlet for each possible combination of the "
                "chosen split_basis.".format(self.name)
            )

        # Get units metadata
        units_meta = self.config.property_package.get_metadata()

        flow_basis = mb[self.flowsheet().time.first()].get_material_flow_basis()
        if flow_basis == MaterialFlowBasis.molar:
            flow_units = units_meta.get_derived_units("flow_mole")
        elif flow_basis == MaterialFlowBasis.mass:
            flow_units = units_meta.get_derived_units("flow_mass")
        else:
            # Let this pass for now with no units
            flow_units = None

        # Create tolerance Parameter for 0 flow outlets
        self.eps = Param(default=1e-8, mutable=True, units=flow_units)

        # Get list of port members
        s_vars = mb[self.flowsheet().time.first()].define_port_members()

        # Get phase component list(s)
        pc_set = mb.phase_component_set

        # Add empty Port objects
        for o in outlet_list:
            p_obj = Port(noruleinit=True, doc="Outlet Port")
            setattr(self, o, p_obj)

            # Iterate over members to create References or Expressions
            for var_name_port, var_obj in s_vars.items():
                # Get local variable name of component
                var_name_local = var_obj.local_name
                if var_name_local == "pressure" or var_name_local == "temperature":
                    # Assume outlets same as mixed flow - make Reference
                    e_obj = Reference(mb[:].component(var_name_local))

                elif var_name_local.startswith(
                    "mole_frac"
                ) or var_name_local.startswith("mass_frac"):
                    # Mole and mass frac need special handling
                    if "_phase" in var_name_local:
                        e_obj = VarLikeExpression(
                            self.flowsheet().time,
                            pc_set,
                            rule=partial(
                                _e_rule_mole_mass_frac_phase_comp,
                                mixed_block=mb,
                                var_obj=var_obj,
                                outlet=o,
                            ),
                        )

                    else:
                        e_obj = VarLikeExpression(
                            self.flowsheet().time,
                            mb.component_list,
                            rule=partial(
                                _e_rule_mole_mass_frac_comp,
                                mixed_block=mb,
                                var_obj=var_obj,
                                outlet=o,
                            ),
                        )

                elif var_name_local.endswith("_phase_comp"):
                    e_obj = VarLikeExpression(
                        self.flowsheet().time,
                        pc_set,
                        rule=partial(
                            _e_rule_phase_comp,
                            mixed_block=mb,
                            var_obj=var_obj,
                            outlet=o,
                        ),
                    )

                elif var_name_local.endswith("_phase"):
                    e_obj = VarLikeExpression(
                        self.flowsheet().time,
                        mb.phase_list,
                        rule=partial(
                            _e_rule_phase, mixed_block=mb, var_obj=var_obj, outlet=o
                        ),
                    )

                elif var_name_local.endswith("_comp"):
                    e_obj = VarLikeExpression(
                        self.flowsheet().time,
                        mb.component_list,
                        rule=partial(
                            _e_rule_comp, mixed_block=mb, var_obj=var_obj, outlet=o
                        ),
                    )

                else:
                    # Variable unindexed by phase or component
                    e_obj = VarLikeExpression(
                        self.flowsheet().time,
                        rule=partial(
                            _e_rule_other, mixed_block=mb, var_obj=var_obj, outlet=o
                        ),
                    )

                # Add Reference/Expression object to Separator model object
                setattr(self, "_" + o + "_" + var_name_local + "_ref", e_obj)

                # Add member to Port object
                p_obj.add(e_obj, var_name_port)

    def model_check(blk):
        """
        This method executes the model_check methods on the associated state
        blocks (if they exist). This method is generally called by a unit model
        as part of the unit's model_check method.

        Args:
            None

        Returns:
            None
        """
        # Try property block model check
        for t in blk.flowsheet().time:
            try:
                if blk.config.mixed_state_block is None:
                    blk.mixed_state[t].model_check()
                else:
                    blk.config.mixed_state_block.model_check()
            except AttributeError:
                _log.warning(
                    "{} Separator inlet state block has no "
                    "model check. To correct this, add a "
                    "model_check method to the associated "
                    "StateBlock class.".format(blk.name)
                )

            try:
                outlet_list = blk.create_outlet_list()
                for o in outlet_list:
                    o_block = getattr(blk, o + "_state")
                    o_block[t].model_check()
            except AttributeError:
                _log.warning(
                    "{} Separator outlet state block has no "
                    "model checks. To correct this, add a model_check"
                    " method to the associated StateBlock class.".format(blk.name)
                )

    def initialize_build(
        blk, outlvl=idaeslog.NOTSET, optarg=None, solver=None, hold_state=False
    ):
        """
        Initialization routine for separator

        Keyword Arguments:
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)
            hold_state : flag indicating whether the initialization routine
                     should unfix any state variables fixed during
                     initialization, **default** - False. **Valid values:**
                     **True** - states variables are not unfixed, and a dict of
                     returned containing flags for which states were fixed
                     during initialization, **False** - state variables are
                     unfixed after initialization by calling the release_state
                     method.

        Returns:
            If hold_states is True, returns a dict containing flags for which
            states were fixed during initialization.
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        # Initialize mixed state block
        if blk.config.mixed_state_block is not None:
            mblock = blk.config.mixed_state_block
        else:
            mblock = blk.mixed_state
        flags = mblock.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True,
        )

        # Solve for split fractions only
        component_status = {}
        for c in blk.component_objects((Block, Constraint)):
            for i in c:
                if not c[i].local_name == "sum_split_frac":
                    # Record current status of components to restore later
                    component_status[c[i]] = c[i].active
                    c[i].deactivate()

        if degrees_of_freedom(blk) != 0:
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = opt.solve(blk, tee=slc.tee)
                init_log.info(
                    "Initialization Step 1 Complete: {}".format(idaeslog.condition(res))
                )

        for c, s in component_status.items():
            if s:
                c.activate()

        if blk.config.ideal_separation:
            # If using ideal splitting, initialization should be complete
            return flags

        # Initialize outlet StateBlocks
        outlet_list = blk.create_outlet_list()

        # Premises for initializing outlet states:
        # 1. Intensive states remain unchanged - this is either a valid premise
        # or the actual state is impossible to calculate without solving the
        # full separator model.
        # 2. Extensive states are use split fractions if index matches, or
        # average of split fractions for outlet otherwise
        for o in outlet_list:
            # Get corresponding outlet StateBlock
            o_block = getattr(blk, o + "_state")

            # Create dict to store fixed status of state variables
            o_flags = {}
            for t in blk.flowsheet().time:

                # Calculate values for state variables
                s_vars = o_block[t].define_state_vars()
                for v in s_vars:
                    for k in s_vars[v]:
                        # Record whether variable was fixed or not
                        o_flags[t, v, k] = s_vars[v][k].fixed

                        # If fixed, use current value
                        # otherwise calculate guess from mixed state and fix
                        if not s_vars[v][k].fixed:
                            m_var = getattr(mblock[t], s_vars[v].local_name)
                            if "flow" in v:
                                # If a "flow" variable, is extensive
                                # Apply split fraction
                                if blk.config.split_basis == SplittingType.totalFlow:
                                    # All flows split by outlet
                                    s_vars[v][k].fix(
                                        value(m_var[k] * blk.split_fraction[(t, o)])
                                    )
                                elif "_phase_comp" in v:
                                    # Need to match indices, but use split frac
                                    if (
                                        blk.config.split_basis
                                        == SplittingType.phaseComponentFlow
                                    ):
                                        s_vars[v][k].fix(
                                            value(
                                                m_var[k]
                                                * blk.split_fraction[(t, o) + (k,)]
                                            )
                                        )
                                    elif (
                                        blk.config.split_basis
                                        == SplittingType.phaseFlow
                                    ):
                                        s_vars[v][k].fix(
                                            value(
                                                m_var[k]
                                                * blk.split_fraction[(t, o) + (k[0],)]
                                            )
                                        )
                                    elif (
                                        blk.config.split_basis
                                        == SplittingType.componentFlow
                                    ):
                                        s_vars[v][k].fix(
                                            value(
                                                m_var[k]
                                                * blk.split_fraction[(t, o) + (k[1],)]
                                            )
                                        )
                                    else:
                                        raise BurntToast(
                                            "{} encountered unrecognised "
                                            "SplittingType. This should not "
                                            "occur - please send this bug to "
                                            "the IDAES developers.".format(blk.name)
                                        )
                                elif "_phase" in v:
                                    if (
                                        blk.config.split_basis
                                        == SplittingType.phaseComponentFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                blk.split_fraction[t, o, k, j]
                                                for j in mblock.component_list
                                            )
                                            / len(mblock.component_list)
                                        )
                                        s_vars[v][k].fix(value(m_var[k] * avg_split))
                                    elif (
                                        blk.config.split_basis
                                        == SplittingType.phaseFlow
                                    ):
                                        s_vars[v][k].fix(
                                            value(
                                                m_var[k]
                                                * blk.split_fraction[(t, o) + (k,)]
                                            )
                                        )
                                    elif (
                                        blk.config.split_basis
                                        == SplittingType.componentFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                blk.split_fraction[t, o, j]
                                                for j in mblock.component_list
                                            )
                                            / len(mblock.component_list)
                                        )
                                        s_vars[v][k].fix(value(m_var[k] * avg_split))
                                    else:
                                        raise BurntToast(
                                            "{} encountered unrecognised "
                                            "SplittingType. This should not "
                                            "occur - please send this bug to "
                                            "the IDAES developers.".format(blk.name)
                                        )
                                elif "_comp" in v:
                                    if (
                                        blk.config.split_basis
                                        == SplittingType.phaseComponentFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                blk.split_fraction[t, o, p, k]
                                                for p in mblock.phase_list
                                            )
                                            / len(mblock.phase_list)
                                        )
                                        s_vars[v][k].fix(value(m_var[k] * avg_split))
                                    elif (
                                        blk.config.split_basis
                                        == SplittingType.phaseFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                blk.split_fraction[t, o, p]
                                                for p in mblock.phase_list
                                            )
                                            / len(mblock.phase_list)
                                        )
                                        s_vars[v][k].fix(value(m_var[k] * avg_split))
                                    elif (
                                        blk.config.split_basis
                                        == SplittingType.componentFlow
                                    ):
                                        s_vars[v][k].fix(
                                            value(
                                                m_var[k]
                                                * blk.split_fraction[(t, o) + (k,)]
                                            )
                                        )
                                    else:
                                        raise BurntToast(
                                            "{} encountered unrecognised "
                                            "SplittingType. This should not "
                                            "occur - please send this bug to "
                                            "the IDAES developers.".format(blk.name)
                                        )
                                else:
                                    # Assume unindexed extensive state
                                    # Need average split
                                    if (
                                        blk.config.split_basis
                                        == SplittingType.phaseComponentFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                blk.split_fraction[t, o, p, j]
                                                for (p, j) in mblock.phase_component_set
                                            )
                                            / len(mblock.phase_component_set)
                                        )
                                    elif (
                                        blk.config.split_basis
                                        == SplittingType.phaseFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                blk.split_fraction[t, o, p]
                                                for p in mblock.phase_list
                                            )
                                            / len(mblock.phase_list)
                                        )
                                    elif (
                                        blk.config.split_basis
                                        == SplittingType.componentFlow
                                    ):
                                        # Need average split fraction
                                        avg_split = value(
                                            sum(
                                                blk.split_fraction[t, o, j]
                                                for j in mblock.component_list
                                            )
                                            / len(mblock.component_list)
                                        )
                                    else:
                                        raise BurntToast(
                                            "{} encountered unrecognised "
                                            "SplittingType. This should not "
                                            "occur - please send this bug to "
                                            "the IDAES developers.".format(blk.name)
                                        )
                                    s_vars[v][k].fix(value(m_var[k] * avg_split))
                            else:
                                # Otherwise intensive, equate to mixed stream
                                s_vars[v][k].fix(m_var[k].value)

            # Call initialization routine for outlet StateBlock
            o_block.initialize(
                outlvl=outlvl,
                optarg=optarg,
                solver=solver,
                hold_state=False,
            )

            # Revert fixed status of variables to what they were before
            for t in blk.flowsheet().time:
                s_vars = o_block[t].define_state_vars()
                for v in s_vars:
                    for k in s_vars[v]:
                        s_vars[v][k].fixed = o_flags[t, v, k]

        if blk.config.mixed_state_block is None:
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = opt.solve(blk, tee=slc.tee)

            if not check_optimal_termination(res):
                raise InitializationError(
                    f"{blk.name} failed to initialize successfully. Please "
                    f"check the output logs for more information."
                )

            init_log.info(
                "Initialization Step 2 Complete: {}".format(idaeslog.condition(res))
            )
        else:
            init_log.info("Initialization Complete.")

        if hold_state is True:
            return flags
        else:
            blk.release_state(flags, outlvl=outlvl)

    def release_state(blk, flags, outlvl=idaeslog.NOTSET):
        """
        Method to release state variables fixed during initialization.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state = True.
            outlvl : sets output level of logging

        Returns:
            None
        """
        if blk.config.mixed_state_block is None:
            mblock = blk.mixed_state
        else:
            mblock = blk.config.mixed_state_block

        mblock.release_state(flags, outlvl=outlvl)

    def calculate_scaling_factors(self):
        mb_type = self.config.material_balance_type
        mixed_state = self.get_mixed_state_block()
        if mb_type == MaterialBalanceType.useDefault:
            t_ref = self.flowsheet().time.first()
            mb_type = mixed_state[t_ref].default_material_balance_type()
        super().calculate_scaling_factors()

        if hasattr(self, "temperature_equality_eqn"):
            for (t, i), c in self.temperature_equality_eqn.items():
                s = iscale.get_scaling_factor(
                    mixed_state[t].temperature, default=1, warning=True
                )
                iscale.constraint_scaling_transform(c, s)

        if hasattr(self, "pressure_equality_eqn"):
            for (t, i), c in self.pressure_equality_eqn.items():
                s = iscale.get_scaling_factor(
                    mixed_state[t].pressure, default=1, warning=True
                )
                iscale.constraint_scaling_transform(c, s)

        if hasattr(self, "material_splitting_eqn"):
            if mb_type == MaterialBalanceType.componentPhase:
                for (t, _, p, j), c in self.material_splitting_eqn.items():
                    flow_term = mixed_state[t].get_material_flow_terms(p, j)
                    s = iscale.get_scaling_factor(flow_term, default=1)
                    iscale.constraint_scaling_transform(c, s)
            elif mb_type == MaterialBalanceType.componentTotal:
                for (t, _, j), c in self.material_splitting_eqn.items():
                    s = None
                    for p in mixed_state.phase_list:
                        try:
                            ft = mixed_state[t].get_material_flow_terms(p, j)
                        except KeyError:
                            # This component does not exist in this phase
                            continue
                        if s is None:
                            s = iscale.get_scaling_factor(ft, default=1)
                        else:
                            _s = iscale.get_scaling_factor(ft, default=1)
                            s = _s if _s < s else s
                    iscale.constraint_scaling_transform(c, s)
            elif mb_type == MaterialBalanceType.total:
                pc_set = mixed_state.phase_component_set
                for (t, _), c in self.material_splitting_eqn.items():
                    for i, (p, j) in enumerate(pc_set):
                        ft = mixed_state[t].get_material_flow_terms(p, j)
                        if i == 0:
                            s = iscale.get_scaling_factor(ft, default=1)
                        else:
                            _s = iscale.get_scaling_factor(ft, default=1)
                            s = _s if _s < s else s
                    iscale.constraint_scaling_transform(c, s)

    def _get_performance_contents(self, time_point=0):
        if hasattr(self, "split_fraction"):
            var_dict = {}
            for k, v in self.split_fraction.items():
                if k[0] == time_point:
                    var_dict[f"Split Fraction [{str(k[1:])}]"] = v
            return {"vars": var_dict}
        else:
            return None

    def _get_stream_table_contents(self, time_point=0):
        outlet_list = self.create_outlet_list()

        if not self.config.ideal_separation:
            io_dict = {}
            if self.config.mixed_state_block is None:
                io_dict["Inlet"] = self.mixed_state
            else:
                io_dict["Inlet"] = self.config.mixed_state_block

            for o in outlet_list:
                io_dict[o] = getattr(self, o + "_state")

            return create_stream_table_dataframe(io_dict, time_point=time_point)

        else:
            stream_attributes = {}
            stream_attributes["Units"] = {}

            for n in ["inlet"] + outlet_list:
                port_obj = getattr(self, n)

                stream_attributes[n] = {}

                for k in port_obj.vars:
                    for i in port_obj.vars[k]:
                        if isinstance(i, float):
                            quant = report_quantity(port_obj.vars[k][time_point])
                            stream_attributes[n][k] = quant.m
                            stream_attributes["Units"][k] = quant.u
                        else:
                            if len(i) == 2:
                                kname = str(i[1])
                            else:
                                kname = str(i[1:])
                            quant = report_quantity(port_obj.vars[k][time_point, i[1:]])
                            stream_attributes[n][k + " " + kname] = quant.m
                            stream_attributes["Units"][k + " " + kname] = quant.u

            return DataFrame.from_dict(stream_attributes, orient="columns")


def _raise_split_indexed_fail_err(name, s):
    raise ConfigurationError(
        "{} Cannot use ideal splitting with "
        "this property package. Package uses "
        "indexed port member {} which cannot "
        "be partitioned. Please set "
        "configuration argument "
        "ideal_separation = False for this "
        "property package.".format(name, s)
    )


def _raise_split_unindexed_fail_err(name, s):
    raise ConfigurationError(
        "{} Cannot use ideal splitting with "
        "this property package. Package uses "
        "unindexed port member {} which does "
        "not have an equivalent indexed form. ".format(name, s)
    )


# Mole and mass frac need special handling
def _e_rule_mole_mass_frac_phase_comp(b, t, p, j, mixed_block, var_obj, outlet):
    pc_set = mixed_block.phase_component_set
    split_map = b.config.ideal_split_map

    if (p, j) in pc_set:
        if b.config.split_basis == SplittingType.phaseFlow:
            return var_obj[p, j]
        elif b.config.split_basis == SplittingType.componentFlow:
            if split_map[j] == outlet:
                return 1
            else:
                return b.eps
        elif b.config.split_basis == SplittingType.phaseComponentFlow:
            for ps in mixed_block.phase_list:
                if split_map[ps, j] == outlet:
                    return 1
            else:  # pylint: disable=W0120
                return b.eps
    else:
        raise BurntToast(
            "{} This should not happen. Please "
            "report this bug to the IDAES "
            "developers.".format(b.name)
        )


def _e_rule_mole_mass_frac_comp(b, t, j, mixed_block, var_obj, outlet):
    pc_set = mixed_block.phase_component_set
    l_name = var_obj.local_name
    split_map = b.config.ideal_split_map

    if b.config.split_basis == SplittingType.componentFlow:
        if split_map[j] == outlet:
            return 1
        # else:
        return b.eps

    elif b.config.split_basis == SplittingType.phaseComponentFlow:
        if any(split_map[phase, j] == outlet for phase in mixed_block.phase_list):
            return 1
        # else:
        return b.eps

    else:
        mfp = mixed_block[t].component(l_name.replace("_comp", "_phase_comp"))

        if mfp is None:
            raise ConfigurationError(
                "{} Cannot use ideal splitting with "
                "this property package. Package uses "
                "indexed port member {} which cannot "
                "be partitioned. Please set "
                "configuration argument "
                "ideal_separation = False for this "
                "property package.".format(b.name, l_name)
            )

        for phase in mixed_block.phase_list:
            if b.config.split_basis == SplittingType.phaseFlow:
                s_check = split_map[phase]
            else:
                raise BurntToast(
                    "{} This should not happen. Please"
                    " report this bug to the IDAES "
                    "developers.".format(b.name)
                )

            if s_check == outlet and (phase, j) in pc_set:
                return mfp[phase, j]
        # else:
        return b.eps


def _e_rule_phase_comp(b, t, p, j, mixed_block, var_obj, outlet):
    l_name = var_obj.local_name
    split_map = b.config.ideal_split_map

    if b.config.split_basis == SplittingType.phaseFlow:
        s_check = split_map[p]
    elif b.config.split_basis == SplittingType.componentFlow:
        s_check = split_map[j]
    elif b.config.split_basis == SplittingType.phaseComponentFlow:
        s_check = split_map[p, j]
    else:
        raise BurntToast(
            "{} This should not happen. Please"
            " report this bug to the IDAES "
            "developers.".format(b.name)
        )

    if s_check == outlet:
        return mixed_block[t].component(l_name)[p, j]
    else:
        return b.eps


def _e_rule_phase(b, t, p, mixed_block, var_obj, outlet):
    l_name = var_obj.local_name
    split_map = b.config.ideal_split_map

    if b.config.split_basis == SplittingType.phaseFlow:
        if split_map[p] == outlet:
            return mixed_block[t].component(l_name)[p]
        else:
            return b.eps

    else:
        mfp = mixed_block[t].component(l_name + "_comp")

        if mfp is None:
            _raise_split_indexed_fail_err(b.name, l_name)

        for j in mixed_block.component_list:
            if b.config.split_basis == SplittingType.componentFlow:
                s_check = split_map[j]
            elif b.config.split_basis == SplittingType.phaseComponentFlow:
                s_check = split_map[p, j]
            else:
                raise BurntToast(
                    "{} This should not happen. Please"
                    " report this bug to the IDAES "
                    "developers.".format(b.name)
                )

            if s_check == outlet:
                return mfp[p, j]
        # else:
        return b.eps


def _e_rule_comp(b, t, j, mixed_block, var_obj, outlet):
    pc_set = mixed_block.phase_component_set
    l_name = var_obj.local_name
    split_map = b.config.ideal_split_map

    if b.config.split_basis == SplittingType.componentFlow:
        if split_map[j] == outlet:
            return mixed_block[t].component(l_name)[j]
        else:
            return b.eps

    else:
        mfp = mixed_block[t].component("{0}_phase{1}".format(l_name[:-5], l_name[-5:]))

        if mfp is None:
            _raise_split_indexed_fail_err(b.name, l_name)

        for p in mixed_block.phase_list:
            if (p, j) in pc_set:
                if b.config.split_basis == SplittingType.phaseFlow:
                    s_check = split_map[p]
                elif b.config.split_basis == SplittingType.phaseComponentFlow:
                    s_check = split_map[p, j]
                else:
                    raise BurntToast(
                        "{} This should not happen. Please"
                        " report this bug to the IDAES "
                        "developers.".format(b.name)
                    )

                if s_check == outlet:
                    return mfp[p, j]
        return b.eps


def _e_rule_other(b, t, mixed_block, var_obj, outlet):
    pc_set = mixed_block.phase_component_set
    l_name = var_obj.local_name
    split_map = b.config.ideal_split_map

    if b.config.split_basis == SplittingType.phaseFlow:
        ivar = mixed_block[t].component(l_name + "_phase")
        if ivar is not None:
            for p in mixed_block.phase_list:
                if split_map[p] == outlet:
                    return ivar[p]
                else:
                    continue
        else:
            ivar = mixed_block[t].component(l_name + "_phase_comp")
            if ivar is not None:
                for p in mixed_block.phase_list:
                    if split_map[p] == outlet:
                        return sum(
                            ivar[p, j]
                            for j in mixed_block.component_list
                            if (p, j) in pc_set
                        )
                    else:
                        continue
            else:
                _raise_split_unindexed_fail_err(b.name, l_name)

    elif b.config.split_basis == SplittingType.componentFlow:
        ivar = mixed_block[t].component(l_name + "_comp")
        if ivar is not None:
            for j in mixed_block.component_list:
                if split_map[j] == outlet:
                    return ivar[j]
                else:
                    continue
        else:
            ivar = mixed_block[t].component(l_name + "_phase_comp")
            if ivar is not None:
                for j in mixed_block.component_list:
                    if split_map[j] == outlet:
                        return sum(
                            ivar[p, j]
                            for p in mixed_block.phase_list
                            if (p, j) in pc_set
                        )
                    else:
                        continue
            else:
                _raise_split_unindexed_fail_err(b.name, l_name)
    elif b.config.split_basis == SplittingType.phaseComponentFlow:
        ivar = mixed_block[t].component(l_name + "_phase_comp")
        if ivar is not None:
            for p in mixed_block.phase_list:
                for j in mixed_block.component_list:
                    if split_map[p, j] == outlet and (p, j) in pc_set:
                        return ivar[p, j]
                    else:
                        continue
        else:
            _raise_split_unindexed_fail_err(b.name, l_name)
    else:
        # Unrecognised split tupe
        raise BurntToast(
            "{} received unrecognised value for "
            "split_basis argument. This should never "
            "happen, so please contact the IDAES "
            "developers with this bug.".format(b.name)
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring

__author__ = "John Eslick, Douglas Allan"

import enum

from pyomo.common.config import ConfigValue, In
import pyomo.environ as pyo


from idaes.core import useDefault
import idaes.core.util.scaling as iscale
from idaes.core.util.exceptions import ConfigurationError, InitializationError
import idaes.core.util.model_statistics as mstat
from idaes.core.util.constants import Constants
import idaes.logger as idaeslog


class CV_Bound(enum.Enum):
    EXTRAPOLATE = 1
    NODE_VALUE = 2


def _set_and_get_attr(obj, name, val):
    setattr(obj, name, val)
    return getattr(obj, name)


def _set_default_factor(c, s):
    """Iterate over an indexed component, and set individual scaling factors
    if there is not an existing scaling factor

    Args:
        c: indexed component to be scaled
        s: scaling factor
    """
    for i in c:
        if iscale.get_scaling_factor(c[i]) is None:
            iscale.set_scaling_factor(c[i], s)


def _set_if_unfixed(v, val):
    """Set a variable's value so long as it isn't fixed

    Args:
        v: scalar variable
        val: value to be assigned to v
    """
    if not v.fixed:
        v.set_value(pyo.value(val))


def _create_if_none(blk, var_name, idx_set, units):
    """Looks in a block's config to see whether or not a variable has been
    supplied. If it has been, creates a Reference to that variable. If it
    has not been supplied, creates a local variable.

    Args:
        blk: Pyomo block
        var_name: variable name as string
        idx_set: index set of variable var. If scalar, pass None
        units: variable units
    """
    attr = getattr(blk.config, var_name)
    if attr is None:
        if idx_set is None:
            setattr(
                blk,
                var_name,
                pyo.Var(
                    initialize=0,
                    units=units,
                ),
            )
        else:
            setattr(
                blk,
                var_name,
                pyo.Var(
                    *idx_set,
                    initialize=0,
                    units=units,
                ),
            )
    else:
        setattr(blk, var_name, pyo.Reference(attr))


def _init_solve_block(blk, solver, log):
    """Checks whether solving a block is a square problem. If not, raise
    InitializationError. If it is, solve block while redirecting output to
    IDAES log. Then check whether solve was successful. If not successful,
    raise InitializationError

    Args:
        blk: Pyomo block to solve
        solver: solver object (NOT name string)
        log: IDAES solver log
    """
    if not mstat.degrees_of_freedom(blk) == 0:
        raise InitializationError(
            f"{blk.name} encountered a nonsquare problem during "
            "initialization. Check whether all parameter values are fixed and "
            "keep in mind that the SOC submodel initialization methods assume "
            "that certain degrees of freedom have been fixed by the cell "
            "initialization method."
        )
    with idaeslog.solver_log(log, idaeslog.DEBUG) as slc:
        results = solver.solve(blk, tee=slc.tee)
    if not pyo.check_optimal_termination(results):
        raise InitializationError(
            f"{blk.name} failed to initialize successfully. Please check "
            f"the output logs for more information."
        )


def _interpolate_channel(iz, ifaces, nodes, faces, phi_func, phi_inlet, opposite_flow):
    """PRIVATE Function: Interpolate faces of control volumes in 1D

    Args:
        iz: index of the face
        ifaces: set of face indexes
        nodes: set of node z locations
        faces: set of face z locations
        phi_func: function that returns an expression for the quantity to be
            interpolated at the node as a function of node index
        phi_inlet: expression for the value of the quantity to be interpolated
            at the channel inlet face (not the inlet to the CV)
        method: interpolation method
        opposite_flow: if True assume velocity is negative

    Returns:
        expression for phi at face[iz]
    """
    # I don't always need these, but it doesn't take long to calculate them
    if not opposite_flow:
        izu = iz - 1  # adjacent node upstream of the face
    else:
        izu = iz  # adjacent node upstream of the face
    if iz == ifaces.first() and not opposite_flow:
        return phi_inlet
    if iz == ifaces.last() and opposite_flow:
        return phi_inlet
    return phi_func(izu)


def _interpolate_2D(
    ic,
    ifaces,
    nodes,
    faces,
    phi_func,
    phi_bound_0,
    phi_bound_1,
    derivative=False,
):
    """PRIVATE Function: Interpolate faces of control volumes in 2D

    Args:
        ic: face index in x or z direction matching specified direction
        ifaces: set of face indexes
        nodes: set of node locations in specified direction
        faces: set of face locations in specified direction
        phi_func: function that returns an expression for the quantity to be
            interpolated at a node center as a function of node index
        phi_bound_0: expression for the value of the quantity to be interpolated
            at the 0 bound
        phi_bound_1: expression for the value of the quantity to be interpolated
            at the 1 bound
        derivative: If True estimate derivative

    Returns:
        expression for phi at face
    """
    # TODO add tests to ensure this function works as designed
    # Also, user beware, use of CV_Bound enums is untested
    icu = ic - 1
    icd = ic
    if ic == ifaces.first():
        if not isinstance(phi_bound_0, CV_Bound):
            return phi_bound_0
        icu = ic + 1
    if ic == ifaces.last():
        if not isinstance(phi_bound_1, CV_Bound):
            return phi_bound_1
        icd = ic - 2
    # For now CDS is the only option
    cu = nodes.at(icu)
    cd = nodes.at(icd)
    if not derivative:
        cf = faces.at(ic)
        lambf = (cd - cf) / (cd - cu)
        return (1 - lambf) * phi_func(icd) + lambf * phi_func(icu)
    else:
        # Since we are doing linear interpolation derivative is the slope
        # between node centers even if they are not evenly spaced
        return (phi_func(icd) - phi_func(icu)) / (cd - cu)


class _SubsetOf(object):
    # Copied and modified from Pyomo's implementation of In
    #  ___________________________________________________________________________
    #
    #  Pyomo: Python Optimization Modeling Objects
    #  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
    #  Under the terms of Contract DE-NA0003525 with National Technology and
    #  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain
    #  rights in this software.
    #  This software is distributed under the 3-clause BSD License.
    #  ___________________________________________________________________________
    #
    #  This module was originally developed as part of the PyUtilib project
    #  Copyright (c) 2008 Sandia Corporation.
    #  This software is distributed under the BSD License.
    #  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
    #  the U.S. Government retains certain rights in this software.
    #  ___________________________________________________________________________
    """_SubsetOf(domain, cast=None)
    Domain validation class admitting a Container of possible values

    This will admit any iterable containing values that are in the
     `domain` Container (i.e., Container.__contains__() returns True).
    Most common domains are list, set, and dict objects.

    Parameters
    ----------
    domain: Container
        The container that specifies the allowable values.  Incoming
        values are passed to ``domain.__contains__()``, and if ``True``
        is returned, the value is accepted and returned.
    """

    #  TODO Need to determine what to do about repeated entries before this can become a Pyomo PR
    def __new__(cls, domain=None, cast=None):
        return super(_SubsetOf, cls).__new__(cls)

    def __init__(self, domain):
        self._domain = domain

    def __call__(self, possible_subset):
        # We often like to use None to represent the empty set in Config blocks because using a mutable object
        # for a default function value results in the same object being used every time the function is called
        # Not sure the same is true for Config blocks, but it probably is.
        if possible_subset is None:
            return None
        for value in possible_subset:
            if value not in self._domain:
                raise ValueError("value %s not in domain %s" % (value, self._domain))
        return possible_subset


_element_list = ["Ar", "H", "O", "C", "N", "S"]
_gas_species_list = [
    "Ar",
    "CH4",
    "CO",
    "CO2",
    "C2H4",
    "C2H6",
    "C3H8",
    "H2",
    "H2O",
    "H2S",
    "NO",
    "N2",
    "N2O",
    "O2",
    "SO2",
]
_all_species_list = _gas_species_list.copy()
_all_species_list.extend(["Vac", "O^2-", "e^-"])
_element_dict = {
    "Ar": {"Ar": 1},
    "H": {"CH4": 4, "C2H4": 4, "C2H6": 6, "C3H8": 8, "H2": 2, "H2O": 2, "H2S": 2},
    "O": {"CO": 1, "CO2": 2, "H2O": 1, "NO": 1, "N2O": 1, "O2": 2, "SO2": 2},
    "C": {"CH4": 1, "CO": 1, "CO2": 1, "C2H4": 4, "C2H6": 2, "C3H8": 3},
    "N": {"NO": 1, "N2": 2, "N2O": 2},
    "S": {"H2S": 1, "SO2": 1},
}

# If a molecule is not listed under an element, it has zero atoms of that element
for value in _element_dict.values():
    for species in _gas_species_list:
        if species not in value.keys():
            value[species] = 0

# Parameters for binary diffusion coefficients from:
#  "Properties of Gases and Liquids" 5th Ed.
bin_diff_sigma = {
    "Ar": 3.542,
    "Air": 3.711,
    "CH4": 3.758,
    "CO": 3.690,
    "CO2": 3.941,
    "C2H4": 4.163,
    "C2H6": 4.443,
    "C3H8": 5.118,
    "H2": 2.827,
    "H2O": 2.641,
    "H2S": 3.623,
    "NO": 3.492,
    "N2": 3.798,
    "N2O": 3.826,
    "O2": 3.467,
    "SO2": 4.112,
}
bin_diff_epsok = {
    "Ar": 93.3,
    "Air": 78.6,
    "CH4": 148.6,
    "CO": 91.7,
    "CO2": 195.2,
    "C2H4": 224.7,
    "C2H6": 215.7,
    "C3H8": 237.1,
    "H2": 59.7,
    "H2O": 809.1,
    "H2S": 301.1,
    "NO": 116.7,
    "N2": 71.4,
    "N2O": 232.4,
    "O2": 106.7,
    "SO2": 335.4,
}
bin_diff_M = {
    "Ar": 39.948,
    "Air": 28.96,
    "CH4": 16.0425,
    "CO": 28.0101,
    "CO2": 44.0095,
    "C2H4": 28.0532,
    "C2H6": 30.0690,
    "C3H8": 44.0956,
    "H2": 2.01588,
    "H2O": 18.0153,
    "H2S": 34.081,
    "NO": 30.0061,
    "N2": 28.0134,
    "N2O": 44.0128,
    "O2": 31.9988,
    "SO2": 64.064,
}

# Shomate equation parameters
# NIST Webbook
h_params = {
    "Ar": {
        "A": 20.78600,
        "B": 2.825911e-7,
        "C": -1.464191e-7,
        "D": 1.092131e-8,
        "E": -3.661371e-8,
        "F": -6.197350,
        "G": 179.9990,
        "H": 0.000000,
    },
    "CH4": {
        "A": -0.703029,
        "B": 108.4773,
        "C": -42.52157,
        "D": 5.862788,
        "E": 0.678565,
        "F": -76.84376,
        "G": 158.7163,
        "H": -74.87310,
    },
    "CO": {
        "A": 25.56759,
        "B": 6.096130,
        "C": 4.054656,
        "D": -2.671301,
        "E": 0.131021,
        "F": -118.0089,
        "G": 227.3665,
        "H": -110.5271,
    },
    "CO2": {
        "A": 24.99735,
        "B": 55.18696,
        "C": -33.69137,
        "D": 7.948387,
        "E": -0.136638,
        "F": -403.6075,
        "G": 228.2431,
        "H": -393.5224,
    },
    "C2H4": {
        "A": -6.387880,
        "B": 184.4019,
        "C": -112.9718,
        "D": 28.49593,
        "E": 0.315540,
        "F": 48.17332,
        "G": 163.1568,
        "H": 52.46694,
    },
    "H2": {  # 1000K-2500K
        "A": 33.066178,
        "B": -11.363417,
        "C": 11.432816,
        "D": -2.772874,
        "E": -0.158558,
        "F": -9.980797,
        "G": 172.707974,
        "H": 0.0,
    },
    "H2O": {  # 500K-1700K
        "A": 30.09200,
        "B": 6.832514,
        "C": 6.793435,
        "D": -2.534480,
        "E": 0.082139,
        "F": -250.8810,
        "G": 223.3967,
        "H": -241.8264,
    },
    "H2S": {
        "A": 26.88412,
        "B": 18.67809,
        "C": 3.434203,
        "D": -3.378702,
        "E": 0.135882,
        "F": -28.91211,
        "G": 233.3747,
        "H": -20.50202,
    },
    "N2": {
        "A": 19.50583,
        "B": 19.88705,
        "C": -8.598535,
        "D": 1.369784,
        "E": 0.527601,
        "F": -4.935202,
        "G": 212.3900,
        "H": 0.0,
    },
    "O2": {  # 700 to 2000
        "A": 30.03235,
        "B": 8.772972,
        "C": -3.988133,
        "D": 0.788313,
        "E": -0.741599,
        "F": -11.32468,
        "G": 236.1663,
        "H": 0.0,
    },
    "SO2": {
        "A": 21.43049,
        "B": 74.35094,
        "C": -57.75217,
        "D": 16.35534,
        "E": 0.086731,
        "F": -305.7688,
        "G": 254.8872,
        "H": -296.8422,
    },
    "Vac": {
        "A": 0.0,
        "B": 0.0,
        "C": 0.0,
        "D": 0.0,
        "E": 0.0,
        "F": 0.0,
        "G": 0.0,
        "H": 0.0,
    },
    "O^2-": {  # Chosen to match O2 coeffs
        "A": 0.5 * 30.03235,
        "B": 0.5 * 8.772972,
        "C": 0.5 * -3.988133,
        "D": 0.5 * 0.788313,
        "E": 0.5 * -0.741599,
        "F": 0.5 * -11.32468,
        "G": 0.5 * 236.1663,  # 0,
        "H": 0,  # -236.0,
    },
    "e^-": {
        "A": 0.0,
        "B": 0.0,
        "C": 0.0,
        "D": 0.0,
        "E": 0.0,
        "F": 0.0,
        "G": 0.0,
        "H": 0.0,
    },
}


def _binary_diffusion_coefficient_expr(temperature, p, c1, c2):
    mab = 2 * (1.0 / bin_diff_M[c1] + 1.0 / bin_diff_M[c2]) ** (-1)
    sab = (bin_diff_sigma[c1] + bin_diff_sigma[c2]) / 2.0
    epsok = (bin_diff_epsok[c1] * bin_diff_epsok[c2]) ** 0.5
    tr = temperature / (epsok * pyo.units.K)
    a = 1.06036
    b = 0.15610
    c = 0.19300
    d = 0.47635
    e = 1.03587
    f = 1.52996
    g = 1.76474
    h = 3.89411
    omega = a / tr**b + c / pyo.exp(d * tr) + e / pyo.exp(f * tr) + g / pyo.exp(h * tr)
    cm2_to_m2 = 0.01 * 0.01
    Pa_to_bar = 1e-5
    return (
        (
            0.00266
            * cm2_to_m2
            * (temperature / pyo.units.K) ** (3 / 2)
            / (p / pyo.units.Pa)
            / Pa_to_bar
            / mab**0.5
            / sab**2
            / omega
        )
        * pyo.units.m**2
        / pyo.units.s
    )


def _pure_component_cp(temperature, comp):
    d = h_params[comp]
    t = temperature / 1000.0 / pyo.units.K
    return (
        (d["A"] + d["B"] * t + d["C"] * t**2 + d["D"] * t**3 + d["E"] / t**2)
        * pyo.units.J
        / pyo.units.mol
        / pyo.units.K
    )


def _pure_component_cv(temperature, comp):
    return _pure_component_cp(temperature, comp) - Constants.gas_constant


def _pure_component_visc(temperature, comp):
    # The properties of gases and liquids 5th ed.  eqn 9-4.6
    # low pressure gas from theory
    return (
        (
            1e-7
            * 16.64
            * bin_diff_M[comp] ** 0.5
            * temperature
            / (bin_diff_epsok[comp] ** 0.5 * bin_diff_sigma[comp] ** 2)
        )
        * pyo.units.Pa
        * pyo.units.s
    )


def _pure_component_thermal_comductivity(temperature, c):
    # The properties of gases and liquids 5th ed.  eqn 10-3.6
    # low pressure gas Stiel and Thodos 1964.
    cv = _pure_component_cv(temperature, c)
    visc = _pure_component_visc(temperature, c)
    m = bin_diff_M[c] / 1000
    units = pyo.units.W / pyo.units.m / pyo.units.K
    return (1.15 + Constants.gas_constant * 2.03 / cv) * visc * cv / m * units


def _comp_enthalpy_expr(temperature, comp):
    # ideal gas enthalpy
    d = h_params[comp]
    t = temperature / 1000.0 / pyo.units.K
    return (
        1000
        * (
            d["A"] * t
            + d["B"] * t**2 / 2.0
            + d["C"] * t**3 / 3.0
            + d["D"] * t**4 / 4.0
            - d["E"] / t
            + d["F"]
        )
        * pyo.units.J
        / pyo.units.mol
    )


_monotomic_gas_standard_state = ["He", "Ne", "Ar", "Kr", "Xe", "Ra"]
_diatomic_gas_standard_state = ["F", "Cl", "H", "N", "O"]


def _comp_int_energy_expr(temperature, comp):
    # ideal gas internal energy
    # NIST has 298.15 K as a reference state, so adjust internal energy expression for that
    T_ref = 298.15 * pyo.units.K
    dn_form = 1
    for element, molecule_dict in _element_dict.items():
        if element in _monotomic_gas_standard_state:
            dn_form -= molecule_dict[comp]
        elif element in _diatomic_gas_standard_state:
            dn_form -= 0.5 * molecule_dict[comp]
    return (
        _comp_enthalpy_expr(temperature, comp)
        - Constants.gas_constant * (temperature - T_ref)
        + dn_form * Constants.gas_constant * T_ref
    )


def _comp_entropy_expr(temperature, comp):
    # ideal gas entropy
    d = h_params[comp]
    t = temperature / 1000.0 / pyo.units.K
    return (
        (
            d["A"] * pyo.log(t)
            + d["B"] * t
            + d["C"] * t**2 / 2.0
            + d["D"] * t**3 / 3.0
            - d["E"] / 2.0 / t**2
            + d["G"]
        )
        * pyo.units.J
        / pyo.units.mol
        / pyo.units.K
    )


def _submodel_boilerplate_config(CONFIG):
    # Add common fields to submodel CONFIG dict
    CONFIG.declare(
        "control_volume_zfaces",
        ConfigValue(
            description="List containing coordinates of control volume faces "
            "in z direction. Coordinates must start with zero, be strictly "
            "increasing, and end with one"
        ),
    )
    CONFIG.declare(
        "length_z",
        ConfigValue(
            default=None, description="Length in the direction of flow (z-direction)"
        ),
    )
    CONFIG.declare(
        "length_y", ConfigValue(default=None, description="Width of cell (y-direction)")
    )
    CONFIG.declare(
        "current_density",
        ConfigValue(default=None, description="Optional current_density variable"),
    )
    CONFIG.declare(
        "include_temperature_x_thermo",
        ConfigValue(
            domain=In([useDefault, True, False]),
            default=True,
            description="Whether to consider temperature variations in "
            "x direction in thermodynamic equations",
        ),
    )


def _submodel_boilerplate_create_if_none(unit):
    # Check for certain common variables on unit block. If they were provided
    # in CONFIG, create references, otherwise create them new on unit block
    tset = unit.flowsheet().config.time
    iznodes = unit.iznodes
    _create_if_none(unit, "length_z", idx_set=None, units=pyo.units.m)
    _create_if_none(unit, "length_y", idx_set=None, units=pyo.units.m)
    _create_if_none(
        unit,
        "current_density",
        idx_set=(tset, iznodes),
        units=pyo.units.A / pyo.units.m**2,
    )


def _thermal_boundary_conditions_config(CONFIG, thin):
    # Add fieldnames relating to thermal boundary conditions to submodel
    # CONFIG dict. If submodel is thin, there is only a single temperature,
    # otherwise there is one at both ends
    CONFIG.declare(
        "temperature_z",
        ConfigValue(
            default=None, description="Temperature as indexed by time " "and z"
        ),
    )
    if thin:
        CONFIG.declare(
            "temperature_deviation_x",
            ConfigValue(
                default=None,
                description="Deviation of temperature " "from temperature_z",
            ),
        )
    else:
        CONFIG.declare(
            "temperature_deviation_x0",
            ConfigValue(
                default=None,
                description="Deviation of temperature at x=0 " "from temperature_z",
            ),
        )
        CONFIG.declare(
            "temperature_deviation_x1",
            ConfigValue(
                default=None,
                description="Deviation of temperature at x=1 " "from temperature_z",
            ),
        )
    CONFIG.declare(
        "heat_flux_x0",
        ConfigValue(
            default=None, description="Heat flux through x=0 " "(positive is in)"
        ),
    )
    CONFIG.declare(
        "heat_flux_x1",
        ConfigValue(
            default=None, description="Heat flux through x=1 " "(positive is out)"
        ),
    )


def _create_thermal_boundary_conditions_if_none(unit, thin):
    # Check for variables relating to thermal boundary conditions on unit
    # block. If they were provided in CONFIG, create references, otherwise
    # create them new on unit block
    tset = unit.flowsheet().config.time
    include_temp_x_thermo = unit.config.include_temperature_x_thermo
    iznodes = unit.iznodes

    _create_if_none(unit, "temperature_z", idx_set=(tset, iznodes), units=pyo.units.K)

    if thin:
        _create_if_none(
            unit, "temperature_deviation_x", idx_set=(tset, iznodes), units=pyo.units.K
        )

        @unit.Expression(tset, iznodes)
        def temperature(b, t, iz):
            if include_temp_x_thermo:
                return b.temperature_z[t, iz] + b.temperature_deviation_x[t, iz]
            else:
                return b.temperature_z[t, iz]

    else:
        _create_if_none(
            unit, "temperature_deviation_x0", idx_set=(tset, iznodes), units=pyo.units.K
        )

        @unit.Expression(tset, iznodes)
        def temperature_x0(b, t, iz):
            if include_temp_x_thermo:
                return b.temperature_z[t, iz] + b.temperature_deviation_x0[t, iz]
            else:
                return b.temperature_z[t, iz]

        _create_if_none(
            unit, "temperature_deviation_x1", idx_set=(tset, iznodes), units=pyo.units.K
        )

        @unit.Expression(tset, iznodes)
        def temperature_x1(b, t, iz):
            if include_temp_x_thermo:
                return b.temperature_z[t, iz] + b.temperature_deviation_x1[t, iz]
            else:
                return b.temperature_z[t, iz]

    _create_if_none(
        unit,
        "heat_flux_x0",
        idx_set=(tset, iznodes),
        units=pyo.units.W / pyo.units.m**2,
    )

    _create_if_none(
        unit,
        "heat_flux_x1",
        idx_set=(tset, iznodes),
        units=pyo.units.W / pyo.units.m**2,
    )


def _material_boundary_conditions_config(CONFIG, thin):
    # Add fieldnames relating to material boundary conditions to submodel
    # CONFIG dict. If submodel is thin, there is only a single concentration,
    # otherwise either end of unit have different concentrations
    if thin:
        CONFIG.declare(
            "material_flux_x",
            ConfigValue(default=None, description="Variable for material flux"),
        )
        CONFIG.declare(
            "conc_mol_comp_deviation_x",
            ConfigValue(
                default=None,
                description="Deviation of concentration from channel bulk "
                "concentration",
            ),
        )
    else:
        CONFIG.declare(
            "conc_mol_comp_deviation_x0",
            ConfigValue(
                default=None,
                description="Deviation of concentration at x= 0 from "
                "channel bulk concentration",
            ),
        )
        CONFIG.declare(
            "conc_mol_comp_deviation_x1",
            ConfigValue(
                default=None,
                description="Deviation of concentration at x= 0 from "
                "channel bulk concentration",
            ),
        )
        CONFIG.declare(
            "material_flux_x0",
            ConfigValue(default=None, description="Variable for material flux at x=0"),
        )
        CONFIG.declare(
            "material_flux_x1",
            ConfigValue(default=None, description="Variable for material flux at x=1"),
        )


def _create_material_boundary_conditions_if_none(unit, thin):
    # Check for variables relating to material boundary conditions on unit
    # block. If they were provided in CONFIG, create references, otherwise
    # create them new on unit block
    tset = unit.flowsheet().config.time
    iznodes = unit.iznodes
    comps = unit.component_list
    if thin:
        _create_if_none(
            unit,
            "conc_mol_comp_deviation_x",
            idx_set=(tset, iznodes, comps),
            units=pyo.units.mol / pyo.units.m**3,
        )
        _create_if_none(
            unit,
            "material_flux_x",
            idx_set=(tset, iznodes, comps),
            units=pyo.units.mol / (pyo.units.s * pyo.units.m**2),
        )
    else:
        _create_if_none(
            unit,
            "conc_mol_comp_deviation_x0",
            idx_set=(tset, iznodes, comps),
            units=pyo.units.mol / pyo.units.m**3,
        )
        _create_if_none(
            unit,
            "conc_mol_comp_deviation_x1",
            idx_set=(tset, iznodes, comps),
            units=pyo.units.mol / pyo.units.m**3,
        )
        _create_if_none(
            unit,
            "material_flux_x0",
            idx_set=(tset, iznodes, comps),
            units=pyo.units.mol / (pyo.units.s * pyo.units.m**2),
        )
        _create_if_none(
            unit,
            "material_flux_x1",
            idx_set=(tset, iznodes, comps),
            units=pyo.units.mol / (pyo.units.s * pyo.units.m**2),
        )


def _face_initializer(blk, faces, direction):
    dfaces = direction + "faces"
    dnodes = direction + "nodes"
    if faces[0] != 0:
        raise ConfigurationError(
            f"Smallest control volume face provided in "
            f"{dfaces} to block {blk.name} is not zero."
        )
    if faces[-1] != 1:
        raise ConfigurationError(
            f"Largest control volume face provided in "
            f"{dfaces} to block {blk.name} is not one."
        )
    for i in range(len(faces) - 1):
        if not faces[i + 1] - faces[i] > 0:
            raise ConfigurationError(
                f"Sequence of control volume face distances in {dfaces} "
                "is not strictly increasing."
            )

    face_set = _set_and_get_attr(
        blk,
        dfaces,
        pyo.Set(
            initialize=faces,
            ordered=True,
            doc=f"{direction} coordinates for control volume faces",
        ),
    )
    node_set = _set_and_get_attr(
        blk,
        dnodes,
        pyo.Set(
            initialize=[
                (face_set.at(i) + face_set.at(i + 1)) / 2.0
                for i in range(1, len(faces))
            ],
            ordered=True,
            doc=f"{direction} coordinates for control volume centers",
        ),
    )
    # This sets provide an integer index for nodes and faces
    iface_set = _set_and_get_attr(
        blk,
        "i" + dfaces,
        pyo.Set(
            initialize=range(1, len(face_set) + 1),
            ordered=True,
            doc=f"Integer index set for {dfaces}",
        ),
    )
    inode_set = _set_and_get_attr(
        blk,
        "i" + dnodes,
        pyo.Set(
            initialize=range(1, len(node_set) + 1),
            ordered=True,
            doc=f"Integer index set for {dnodes}",
        ),
    )
    return iface_set, inode_set



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for math util methods.
"""

import pytest
from pyomo.environ import (
    Block,
    ConcreteModel,
    Constraint,
    Expression,
    exp,
    Set,
    Var,
    value,
    Param,
    Reals,
    units as pyunits,
    TransformationFactory,
    check_optimal_termination,
)
from pyomo.network import Arc, Port

from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    declare_process_block_class,
    PhysicalParameterBlock,
    StateBlock,
    StateBlockData,
    ReactionParameterBlock,
    ReactionBlockBase,
    ReactionBlockDataBase,
    MaterialFlowBasis,
    Component,
    Phase,
)
from idaes.core.util.testing import PhysicalParameterTestBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.models.unit_models import CSTR
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.initialization import (
    fix_state_vars,
    revert_state_vars,
    propagate_state,
    solve_indexed_blocks,
    initialize_by_time_element,
)
from idaes.core.solvers import get_solver

__author__ = "Andrew Lee"


# Set up solver
solver = get_solver("ipopt_v2")


@declare_process_block_class("AqueousEnzymeParameterBlock")
class ParameterData(PhysicalParameterBlock):
    """
    Parameter block for the aqueous enzyme reaction in the biochemical CSTR
    used by Christofides and Daoutidis, 1996, presented by Heineken et al, 1967
    """

    def build(self):
        super(ParameterData, self).build()

        # all components are in the aqueous phase
        self.aq = Phase()
        self.S = Component()
        self.E = Component()
        self.C = Component()
        self.P = Component()
        self.Solvent = Component()

        self._state_block_class = AqueousEnzymeStateBlock

    @classmethod
    def define_metadata(cls, obj):
        obj.add_default_units(
            {
                "time": pyunits.minute,
                "length": pyunits.m,
                "amount": pyunits.kmol,
                "temperature": pyunits.K,
                "mass": pyunits.kg,
            }
        )


class _AqueousEnzymeStateBlock(StateBlock):
    def initialize(blk):
        pass


@declare_process_block_class(
    "AqueousEnzymeStateBlock", block_class=_AqueousEnzymeStateBlock
)
class AqueousEnzymeStateBlockData(StateBlockData):
    def build(self):
        super(AqueousEnzymeStateBlockData, self).build()

        self.conc_mol = Var(
            self._params.component_list,
            domain=Reals,
            doc="Component molar concentration [kmol/m^3]",
        )

        self.flow_mol_comp = Var(
            self._params.component_list,
            domain=Reals,
            doc="Molar component flow rate [kmol/min]",
        )

        self.flow_vol = Var(
            domain=Reals, doc="Volumetric flow rate out of reactor [m^3/min]"
        )

        self.temperature = Var(
            initialize=303, domain=Reals, doc="Temperature within reactor [K]"
        )

        if not self.config.defined_state:
            self.conc_mol["Solvent"].fix(1.0)

        def flow_mol_comp_rule(b, j):
            return b.flow_mol_comp[j] == b.flow_vol * b.conc_mol[j]

        self.flow_mol_comp_eqn = Constraint(
            self._params.component_list,
            rule=flow_mol_comp_rule,
            doc="Outlet component molar flow rate equation",
        )

    def get_material_density_terms(b, p, j):
        return b.conc_mol[j]

    def get_material_flow_terms(b, p, j):
        return b.flow_mol_comp[j]

    def get_material_flow_basis(b):
        return MaterialFlowBasis.molar

    def get_enthalpy_flow_terms(b, p):
        return b.flow_vol * b.temperature

    def get_energy_density_terms(b, p):
        return b.temperature

    def define_state_vars(b):
        return {
            "conc_mol": b.conc_mol,
            "temperature": b.temperature,
            "flow_vol": b.flow_vol,
        }


@declare_process_block_class("EnzymeReactionParameterBlock")
class EnzymeReactionParameterData(ReactionParameterBlock):
    """
    Enzyme reaction:
    S + E <-> C -> P + E
    """

    def build(self):
        super(EnzymeReactionParameterData, self).build()

        self._reaction_block_class = EnzymeReactionBlock

        self.rate_reaction_idx = Set(initialize=["R1", "R2", "R3"])
        self.rate_reaction_stoichiometry = {
            ("R1", "aq", "S"): -1,
            ("R1", "aq", "E"): -1,
            ("R1", "aq", "C"): 1,
            ("R1", "aq", "P"): 0,
            ("R1", "aq", "Solvent"): 0,
            ("R2", "aq", "S"): 1,
            ("R2", "aq", "E"): 1,
            ("R2", "aq", "C"): -1,
            ("R2", "aq", "P"): 0,
            ("R2", "aq", "Solvent"): 0,
            ("R3", "aq", "S"): 0,
            ("R3", "aq", "E"): 1,
            ("R3", "aq", "C"): -1,
            ("R3", "aq", "P"): 1,
            ("R3", "aq", "Solvent"): 0,
        }

        self.act_energy = Param(
            self.rate_reaction_idx,
            initialize={"R1": 8.0e3, "R2": 9.0e3, "R3": 1.0e4},
            doc="Activation energy [kcal/kmol]",
        )

        self.gas_const = Param(initialize=1.987, doc="Gas constant R [kcal/kmol/K]")

        self.temperature_ref = Param(initialize=300.0, doc="Reference temperature")

        self.k_rxn = Param(
            self.rate_reaction_idx,
            initialize={"R1": 3.36e6, "R2": 1.80e6, "R3": 5.79e7},
            doc="Pre-exponential rate constant in Arrhenius expression",
        )

    #    self.reaction_block_class = EnzymeReactionBlock

    @classmethod
    def define_metadata(cls, obj):
        obj.add_default_units(
            {
                "time": pyunits.minute,
                "length": pyunits.m,
                "amount": pyunits.kmol,
                "temperature": pyunits.K,
                "mass": pyunits.kg,
            }
        )


class _EnzymeReactionBlock(ReactionBlockBase):
    def initialize(blk):
        # initialize for reaction rates for each data object
        pass


@declare_process_block_class("EnzymeReactionBlock", block_class=_EnzymeReactionBlock)
class EnzymeReactionBlockData(ReactionBlockDataBase):
    def build(self):
        super(EnzymeReactionBlockData, self).build()

        self.reaction_coef = Var(
            self._params.rate_reaction_idx,
            domain=Reals,
            doc="Reaction rate coefficient",
        )

        self.reaction_rate = Var(
            self._params.rate_reaction_idx,
            domain=Reals,
            doc="Reaction rate [kmol/m^3/min]",
        )

        self.dh_rxn = Param(
            self._params.rate_reaction_idx,
            domain=Reals,
            doc="Heat of reaction",
            initialize={
                "R1": 1e3 / 900 / 0.231,
                "R2": 1e3 / 900 / 0.231,
                "R3": 5e3 / 900 / 0.231,
            },
        )

        def reaction_rate_rule(b, r):
            if r == "R1":
                return (
                    b.reaction_rate[r]
                    == b.reaction_coef[r]
                    * b.state_ref.conc_mol["S"]
                    * b.state_ref.conc_mol["E"]
                )
            elif r == "R2":
                return (
                    b.reaction_rate[r] == b.reaction_coef[r] * b.state_ref.conc_mol["C"]
                )
            elif r == "R3":
                return (
                    b.reaction_rate[r] == b.reaction_coef[r] * b.state_ref.conc_mol["C"]
                )

        self.reaction_rate_eqn = Constraint(
            self._params.rate_reaction_idx, rule=reaction_rate_rule
        )

        def arrhenius_rule(b, r):
            return b.reaction_coef[r] == b._params.k_rxn[r] * exp(
                -b._params.act_energy[r] / b._params.gas_const / b.state_ref.temperature
            )

        self.arrhenius_eqn = Constraint(
            self._params.rate_reaction_idx, rule=arrhenius_rule
        )

    def get_reaction_rate_basis(b):
        return MaterialFlowBasis.molar


@pytest.fixture
def model():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.pp = PhysicalParameterTestBlock()
    m.fs.sb = m.fs.pp.state_block_class(parameters=m.fs.pp)

    for i in m.fs.sb.flow_mol_phase_comp:
        assert not m.fs.sb.flow_mol_phase_comp[i].fixed
        assert m.fs.sb.flow_mol_phase_comp[i].value == 2
    assert not m.fs.sb.pressure.fixed
    assert m.fs.sb.pressure.value == 1e5
    assert not m.fs.sb.temperature.fixed
    assert m.fs.sb.temperature.value == 300

    return m


@pytest.mark.unit
def test_fix_state_vars_basic(model):
    flags = fix_state_vars(model.fs.sb)

    for i in model.fs.sb.flow_mol_phase_comp:
        assert model.fs.sb.flow_mol_phase_comp[i].fixed
        assert model.fs.sb.flow_mol_phase_comp[i].value == 2
    assert model.fs.sb.pressure.fixed
    assert model.fs.sb.pressure.value == 1e5
    assert model.fs.sb.temperature.fixed
    assert model.fs.sb.temperature.value == 300

    assert not flags[None, "component_flow_phase", ("p1", "c1")]
    assert not flags[None, "component_flow_phase", ("p1", "c2")]
    assert not flags[None, "component_flow_phase", ("p2", "c1")]
    assert not flags[None, "component_flow_phase", ("p2", "c2")]
    assert not flags[None, "pressure", None]
    assert not flags[None, "temperature", None]


@pytest.mark.unit
def test_fix_state_vars_None_value(model):
    model.fs.sb.pressure.value = None

    with pytest.raises(ConfigurationError):
        fix_state_vars(model.fs.sb)


@pytest.mark.unit
def test_fix_state_vars_guesses(model):
    # Note that flow_mol_phase_comp is labeled as component_flow
    # in define_state_vars
    state_args = {
        "component_flow_phase": {
            ("p1", "c1"): 1,
            ("p1", "c2"): 2,
            ("p2", "c1"): 3,
            ("p2", "c2"): 4,
        },
        "pressure": 2e5,
        "temperature": 500,
    }

    flags = fix_state_vars(model.fs.sb, state_args)

    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].value == 1
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].value == 2
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].value == 3
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].value == 4

    assert model.fs.sb.pressure.fixed
    assert model.fs.sb.pressure.value == 2e5
    assert model.fs.sb.temperature.fixed
    assert model.fs.sb.temperature.value == 500

    assert not flags[None, "component_flow_phase", ("p1", "c1")]
    assert not flags[None, "component_flow_phase", ("p1", "c2")]
    assert not flags[None, "component_flow_phase", ("p2", "c1")]
    assert not flags[None, "component_flow_phase", ("p2", "c2")]
    assert not flags[None, "pressure", None]
    assert not flags[None, "temperature", None]


@pytest.mark.unit
def test_fix_state_vars_partial_guesses(model):
    # Note that flow_mol_phase_comp is labeled as component_flow
    # in define_state_vars
    state_args = {
        "component_flow_phase": {
            ("p1", "c1"): 1,
            ("p1", "c2"): 2,
            ("p2", "c1"): 3,
            ("p2", "c2"): 4,
        }
    }

    flags = fix_state_vars(model.fs.sb, state_args)

    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].value == 1
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].value == 2
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].value == 3
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].value == 4

    assert model.fs.sb.pressure.fixed
    assert model.fs.sb.pressure.value == 1e5
    assert model.fs.sb.temperature.fixed
    assert model.fs.sb.temperature.value == 300

    assert not flags[None, "component_flow_phase", ("p1", "c1")]
    assert not flags[None, "component_flow_phase", ("p1", "c2")]
    assert not flags[None, "component_flow_phase", ("p2", "c1")]
    assert not flags[None, "component_flow_phase", ("p2", "c2")]
    assert not flags[None, "pressure", None]
    assert not flags[None, "temperature", None]


@pytest.mark.unit
def test_fix_state_vars_guesses_mismatch_index(model):
    # Note that flow_mol_phase_comp is labeled as component_flow
    # in define_state_vars
    state_args = {
        "component_flow_phase": {("p1", "c1"): 1, ("p1", "c2"): 2, ("p2", "c1"): 3},
        "pressure": 2e5,
        "temperature": 500,
    }

    with pytest.raises(ConfigurationError):
        fix_state_vars(model.fs.sb, state_args)


@pytest.mark.unit
def test_fix_state_vars_fixed_no_guesses(model):
    # Note that flow_mol_phase_comp is labeled as component_flow
    # in define_state_vars
    model.fs.sb.flow_mol_phase_comp["p1", "c1"].fix(10)
    model.fs.sb.pressure.fix(1.5e5)

    flags = fix_state_vars(model.fs.sb)

    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].value == 10
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].value == 2
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].value == 2
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].value == 2

    assert model.fs.sb.pressure.fixed
    assert model.fs.sb.pressure.value == 1.5e5
    assert model.fs.sb.temperature.fixed
    assert model.fs.sb.temperature.value == 300

    # Pressure and component_flow_phase[p1, c1] should be True
    assert flags[None, "component_flow_phase", ("p1", "c1")]
    assert not flags[None, "component_flow_phase", ("p1", "c2")]
    assert not flags[None, "component_flow_phase", ("p2", "c1")]
    assert not flags[None, "component_flow_phase", ("p2", "c2")]
    assert flags[None, "pressure", None]
    assert not flags[None, "temperature", None]


@pytest.mark.unit
def test_fix_state_vars_fixed_guesses(model):
    # Note that flow_mol_phase_comp is labeled as component_flow
    # in define_state_vars
    model.fs.sb.flow_mol_phase_comp["p1", "c1"].fix(10)
    model.fs.sb.pressure.fix(1.5e5)

    state_args = {
        "component_flow_phase": {
            ("p1", "c1"): 1,
            ("p1", "c2"): 2,
            ("p2", "c1"): 3,
            ("p2", "c2"): 4,
        },
        "pressure": 2e5,
        "temperature": 500,
    }

    flags = fix_state_vars(model.fs.sb, state_args)

    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].value == 10
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].value == 2
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].value == 3
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].value == 4

    assert model.fs.sb.pressure.fixed
    assert model.fs.sb.pressure.value == 1.5e5
    assert model.fs.sb.temperature.fixed
    assert model.fs.sb.temperature.value == 500

    # Pressure and component_flow_phase[p1, c1] should be True
    assert flags[None, "component_flow_phase", ("p1", "c1")]
    assert not flags[None, "component_flow_phase", ("p1", "c2")]
    assert not flags[None, "component_flow_phase", ("p2", "c1")]
    assert not flags[None, "component_flow_phase", ("p2", "c2")]
    assert flags[None, "pressure", None]
    assert not flags[None, "temperature", None]


@pytest.mark.unit
def test_revert_state_vars_basic(model):
    flags = fix_state_vars(model.fs.sb)

    revert_state_vars(model.fs.sb, flags)

    for i in model.fs.sb.flow_mol_phase_comp:
        assert not model.fs.sb.flow_mol_phase_comp[i].fixed
        assert model.fs.sb.flow_mol_phase_comp[i].value == 2
    assert not model.fs.sb.pressure.fixed
    assert model.fs.sb.pressure.value == 1e5
    assert not model.fs.sb.temperature.fixed
    assert model.fs.sb.temperature.value == 300


@pytest.mark.unit
def test_revert_state_vars_guesses(model):
    # Note that flow_mol_phase_comp is labeled as component_flow
    # in define_state_vars
    state_args = {
        "component_flow_phase": {
            ("p1", "c1"): 1,
            ("p1", "c2"): 2,
            ("p2", "c1"): 3,
            ("p2", "c2"): 4,
        },
        "pressure": 2e5,
        "temperature": 500,
    }

    flags = fix_state_vars(model.fs.sb, state_args)

    revert_state_vars(model.fs.sb, flags)

    assert not model.fs.sb.flow_mol_phase_comp[("p1", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].value == 1
    assert not model.fs.sb.flow_mol_phase_comp[("p1", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].value == 2
    assert not model.fs.sb.flow_mol_phase_comp[("p2", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].value == 3
    assert not model.fs.sb.flow_mol_phase_comp[("p2", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].value == 4

    assert not model.fs.sb.pressure.fixed
    assert model.fs.sb.pressure.value == 2e5
    assert not model.fs.sb.temperature.fixed
    assert model.fs.sb.temperature.value == 500


@pytest.mark.unit
def test_revert_state_vars_fixed_no_guesses(model):
    # Note that flow_mol_phase_comp is labeled as component_flow
    # in define_state_vars
    model.fs.sb.flow_mol_phase_comp["p1", "c1"].fix(10)
    model.fs.sb.pressure.fix(1.5e5)

    flags = fix_state_vars(model.fs.sb)
    revert_state_vars(model.fs.sb, flags)

    # Pressure and component_flow[p1, c1] should still be fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].value == 10
    assert not model.fs.sb.flow_mol_phase_comp[("p1", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].value == 2
    assert not model.fs.sb.flow_mol_phase_comp[("p2", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].value == 2
    assert not model.fs.sb.flow_mol_phase_comp[("p2", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].value == 2

    assert model.fs.sb.pressure.fixed
    assert model.fs.sb.pressure.value == 1.5e5
    assert not model.fs.sb.temperature.fixed
    assert model.fs.sb.temperature.value == 300


@pytest.mark.unit
def test_revert_state_vars_fixed_guesses(model):
    # Note that flow_mol_phase_comp is labeled as component_flow
    # in define_state_vars
    model.fs.sb.flow_mol_phase_comp["p1", "c1"].fix(10)
    model.fs.sb.pressure.fix(1.5e5)

    state_args = {
        "component_flow_phase": {
            ("p1", "c1"): 1,
            ("p1", "c2"): 2,
            ("p2", "c1"): 3,
            ("p2", "c2"): 4,
        },
        "pressure": 2e5,
        "temperature": 500,
    }

    flags = fix_state_vars(model.fs.sb, state_args)
    revert_state_vars(model.fs.sb, flags)

    # Pressure and component_flow[p1, c1] should still be fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c1")].value == 10
    assert not model.fs.sb.flow_mol_phase_comp[("p1", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p1", "c2")].value == 2
    assert not model.fs.sb.flow_mol_phase_comp[("p2", "c1")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c1")].value == 3
    assert not model.fs.sb.flow_mol_phase_comp[("p2", "c2")].fixed
    assert model.fs.sb.flow_mol_phase_comp[("p2", "c2")].value == 4

    assert model.fs.sb.pressure.fixed
    assert model.fs.sb.pressure.value == 1.5e5
    assert not model.fs.sb.temperature.fixed
    assert model.fs.sb.temperature.value == 500


@pytest.mark.unit
def test_revert_state_vars_flag_mismatch(model):
    flags = {
        (None, "component_flow_phase", ("p1", "c1")): False,
        (None, "component_flow_phase", ("p1", "c2")): False,
        (None, "component_flow_phase", ("p2", "c1")): False,
        (None, "pressure", None): False,
    }

    with pytest.raises(ConfigurationError):
        revert_state_vars(model.fs.sb, flags)


@pytest.mark.unit
def test_propagate_state():
    m = ConcreteModel()

    def block_rule(b):
        b.s = Set(initialize=[1, 2])
        b.v1 = Var()
        b.v2 = Var(b.s)
        b.e1 = Expression(expr=b.v1)

        @b.Expression(b.s)
        def e2(blk, i):
            return b.v2[i] * b.v1

        b.p1 = Param(mutable=True, initialize=5)
        b.p2 = Param(b.s, mutable=True, initialize=6)

        b.port1 = Port()
        b.port1.add(b.v1, "V1")
        b.port1.add(b.v2, "V2")

        b.port2 = Port()
        b.port2.add(b.v1, "V1")
        b.port2.add(b.e2, "V2")

        b.port3 = Port()
        b.port3.add(b.e1, "V1")
        b.port3.add(b.v2, "V2")

        b.port4 = Port()
        b.port4.add(b.p1, "V1")
        b.port4.add(b.v2, "V2")

        b.port5 = Port()
        b.port5.add(b.v1, "V1")
        b.port5.add(b.p2, "V2")

        b.port6 = Port()
        b.port6.add(b.v1, "V1")
        b.port6.add(b.v1, "V2")
        return

    m.b1 = Block(rule=block_rule)
    m.b2 = Block(rule=block_rule)

    m.s1 = Arc(source=m.b1.port1, destination=m.b2.port1)
    m.s2 = Arc(source=m.b1.port1, destination=m.b2.port2)
    m.s3 = Arc(source=m.b1.port1, destination=m.b2.port3)
    m.s4 = Arc(source=m.b1.port1, destination=m.b2.port4)
    m.s5 = Arc(source=m.b1.port1, destination=m.b2.port5)
    m.s6 = Arc(source=m.b1.port2, destination=m.b2.port1)
    m.s7 = Arc(source=m.b1.port3, destination=m.b2.port1)
    m.s8 = Arc(source=m.b1.port4, destination=m.b2.port1)
    m.s9 = Arc(source=m.b1.port5, destination=m.b2.port1)
    m.s10 = Arc(source=m.b1.port6, destination=m.b2.port1)
    m.s11 = Arc(source=m.b2.port6, destination=m.b1.port1)

    # Set values on first block
    m.b1.v1.value = 10
    m.b1.v2[1].value = 20
    m.b1.v2[2].value = 30

    # Make sure vars in block 2 haven't been changed accidentally
    assert m.b2.v1.value is None
    assert m.b2.v2[1].value is None
    assert m.b2.v2[2].value is None

    propagate_state(m.s1)

    # Check that values were propagated correctly
    assert m.b2.v1.value == m.b1.v1.value
    assert m.b2.v2[1].value == m.b1.v2[1].value
    assert m.b2.v2[2].value == m.b1.v2[2].value

    assert m.b1.v1.fixed is False
    assert m.b1.v2[1].fixed is False
    assert m.b1.v2[2].fixed is False
    assert m.b2.v1.fixed is False
    assert m.b2.v2[1].fixed is False
    assert m.b2.v2[2].fixed is False

    with pytest.raises(TypeError):
        propagate_state(m.s2)

    with pytest.raises(TypeError):
        propagate_state(m.s3)

    with pytest.raises(TypeError):
        propagate_state(m.s4)

    with pytest.raises(TypeError):
        propagate_state(m.s5)

    propagate_state(m.s6)
    assert value(m.b1.v1) == value(m.b2.v1)
    assert value(m.b1.e2[1]) == value(m.b2.v2[1])
    assert value(m.b1.e2[2]) == value(m.b2.v2[2])

    propagate_state(m.s7)
    assert value(m.b1.e1) == value(m.b2.v1)
    assert value(m.b1.v2[1]) == value(m.b2.v2[1])
    assert value(m.b1.v2[2]) == value(m.b2.v2[2])

    propagate_state(m.s8)
    assert value(m.b1.p1) == value(m.b2.v1)
    assert value(m.b1.v2[1]) == value(m.b2.v2[1])
    assert value(m.b1.v2[2]) == value(m.b2.v2[2])

    propagate_state(m.s9)
    assert value(m.b1.v1) == value(m.b2.v1)
    assert value(m.b1.p2[1]) == value(m.b2.v2[1])
    assert value(m.b1.p2[2]) == value(m.b2.v2[2])

    with pytest.raises(KeyError):
        propagate_state(m.s10)

    with pytest.raises(KeyError):
        propagate_state(m.s11)


@pytest.mark.unit
def test_propagate_state_reverse():
    m = ConcreteModel()

    def block_rule(b):
        b.s = Set(initialize=[1, 2])
        b.v1 = Var()
        b.v2 = Var(b.s)

        b.p = Port()
        b.p.add(b.v1, "V1")
        b.p.add(b.v2, "V2")
        return

    m.b1 = Block(rule=block_rule)
    m.b2 = Block(rule=block_rule)

    m.s1 = Arc(source=m.b1.p, destination=m.b2.p)

    # Test reverse propagation - set values on second block
    m.b2.v1.value = 100
    m.b2.v2[1].value = 200
    m.b2.v2[2].value = 300

    # Make sure vars in block 1 haven't been changed accidentally
    assert m.b1.v1.value is None
    assert m.b1.v2[1].value is None
    assert m.b1.v2[2].value is None

    propagate_state(m.s1, direction="backward")

    # Check that values were propagated correctly
    assert m.b2.v1.value == m.b1.v1.value
    assert m.b2.v2[1].value == m.b1.v2[1].value
    assert m.b2.v2[2].value == m.b1.v2[2].value

    assert m.b1.v1.fixed is False
    assert m.b1.v2[1].fixed is False
    assert m.b1.v2[2].fixed is False
    assert m.b2.v1.fixed is False
    assert m.b2.v2[1].fixed is False
    assert m.b2.v2[2].fixed is False


@pytest.mark.unit
def test_propagate_state_indexed_fixed():
    m = ConcreteModel()

    def block_rule(b):
        b.s = Set(initialize=[1, 2])
        b.v1 = Var()
        b.v2 = Var(b.s)

        b.p = Port(b.s)
        b.p[1].add(b.v1, "V1")
        b.p[2].add(b.v2, "V2")
        return

    m.b1 = Block(rule=block_rule)
    m.b2 = Block(rule=block_rule)

    def arc_rule(m, i):
        return {"source": m.b1.p[i], "destination": m.b2.p[i]}

    m.s1 = Arc([1, 2], rule=arc_rule)

    # Set values on first block
    m.b1.v1.value = 10
    m.b1.v2[1].value = 20
    m.b1.v2[2].value = 30

    # Make sure vars in block 2 haven't been changed accidentally
    assert m.b2.v1.value is None
    assert m.b2.v2[1].value is None
    assert m.b2.v2[2].value is None

    # Fix v1 in block 2
    m.b2.v1.fix(500)

    propagate_state(m.s1[1])

    # Check that values were propagated correctly
    assert m.b2.v1.value == 500
    assert m.b1.v1.fixed is False
    assert m.b2.v1.fixed is True

    propagate_state(m.s1[1], overwrite_fixed=True)

    # Check that values were propagated correctly
    assert m.b2.v1.value == 10
    assert m.b1.v1.fixed is False
    assert m.b2.v1.fixed is True

    propagate_state(m.s1[2])

    # Check that values were propagated correctly
    assert m.b2.v2[1].value == m.b1.v2[1].value
    assert m.b2.v2[2].value == m.b1.v2[2].value
    assert m.b1.v2[1].fixed is False
    assert m.b1.v2[2].fixed is False
    assert m.b2.v2[1].fixed is False
    assert m.b2.v2[2].fixed is False


@pytest.mark.unit
def test_propagate_state_indexed():
    m = ConcreteModel()

    def block_rule(b):
        b.s = Set(initialize=[1, 2])
        b.v1 = Var()
        b.v2 = Var(b.s)

        b.p = Port(b.s)
        b.p[1].add(b.v1, "V1")
        b.p[2].add(b.v2, "V2")
        return

    m.b1 = Block(rule=block_rule)
    m.b2 = Block(rule=block_rule)

    def arc_rule(m, i):
        return {"source": m.b1.p[i], "destination": m.b2.p[i]}

    m.s1 = Arc([1, 2], rule=arc_rule)

    with pytest.raises(AttributeError):
        propagate_state(m.s1)


@pytest.mark.unit
def test_propagate_state_Expression():
    m = ConcreteModel()

    def block_rule(b):
        b.s = Set(initialize=[1, 2])
        b.v1 = Var()
        b.v2 = Var(b.s)

        b.e = Expression(expr=b.v1)

        b.p = Port()
        b.p.add(b.e, "E")
        b.p.add(b.v2, "V2")
        return

    m.b1 = Block(rule=block_rule)
    m.b2 = Block(rule=block_rule)

    m.s1 = Arc(source=m.b1.p, destination=m.b2.p)

    with pytest.raises(TypeError):
        propagate_state(m.s1)


@pytest.mark.unit
def test_propagate_state_invalid_stream():
    m = ConcreteModel()

    with pytest.raises(RuntimeError):
        propagate_state(m)


@pytest.mark.unit
def test_propagate_state_invalid_direction():
    m = ConcreteModel()

    def block_rule(b):
        b.s = Set(initialize=[1, 2])
        b.v1 = Var()
        b.v2 = Var(b.s)

        b.p = Port()
        b.p.add(b.v1, "V1")
        b.p.add(b.v2, "V2")
        return

    m.b1 = Block(rule=block_rule)
    m.b2 = Block(rule=block_rule)

    m.s1 = Arc(source=m.b1.p, destination=m.b2.p)

    with pytest.raises(ValueError):
        propagate_state(m.s1, direction="foo")


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.unit
def test_solve_indexed_block_list():
    # Create an indexed block and try to solve it
    m = ConcreteModel()
    m.s = Set(initialize=[1, 2, 3])

    def block_rule(b, x):
        b.v = Var(initialize=1.0)
        b.c = Constraint(expr=b.v == 2.0)

    m.b = Block(m.s, rule=block_rule)

    solve_indexed_blocks(solver=solver, blocks=[m.b])

    for i in m.s:
        assert value(m.b[i].v == 2.0)


@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.unit
def test_solve_indexed_block_IndexedBlock():
    # Create an indexed block and try to solve it
    m = ConcreteModel()
    m.s = Set(initialize=[1, 2, 3])

    def block_rule(b, x):
        b.v = Var(initialize=1.0)
        b.c = Constraint(expr=b.v == 2.0)

    m.b = Block(m.s, rule=block_rule)

    solve_indexed_blocks(solver=solver, blocks=m.b)

    for i in m.s:
        assert value(m.b[i].v == 2.0)


@pytest.mark.unit
def test_solve_indexed_block_error():
    # Try solve_indexed_block on non-block object
    with pytest.raises(TypeError):
        solve_indexed_blocks(solver=None, blocks=[1, 2, 3])


@pytest.mark.integration
@pytest.mark.skipif(solver is None, reason="Solver not available")
def test_initialize_by_time_element():
    horizon = 6
    time_set = [0, horizon]
    ntfe = 60  # For a finite element every six seconds
    ntcp = 2
    m = ConcreteModel(name="CSTR model for testing")
    m.fs = FlowsheetBlock(dynamic=True, time_set=time_set, time_units=pyunits.minute)

    m.fs.properties = AqueousEnzymeParameterBlock()
    m.fs.reactions = EnzymeReactionParameterBlock(property_package=m.fs.properties)
    m.fs.cstr = CSTR(
        property_package=m.fs.properties,
        reaction_package=m.fs.reactions,
        material_balance_type=MaterialBalanceType.componentTotal,
        energy_balance_type=EnergyBalanceType.enthalpyTotal,
        momentum_balance_type=MomentumBalanceType.none,
        has_heat_of_reaction=True,
    )

    # Time discretization
    disc = TransformationFactory("dae.collocation")
    disc.apply_to(m, wrt=m.fs.time, nfe=ntfe, ncp=ntcp, scheme="LAGRANGE-RADAU")

    # Fix geometry variables
    m.fs.cstr.volume[0].fix(1.0)

    # Fix initial conditions:
    for p, j in m.fs.properties.phase_list * m.fs.properties.component_list:
        if j == "Solvent":
            continue
        m.fs.cstr.control_volume.material_holdup[0, p, j].fix(0)

    # Fix inlet conditions
    # This is a huge hack because I didn't know that the proper way to
    # have multiple inlets to a CSTR was to use a mixer.
    # I 'combine' both my inlet streams before sending them to the CSTR.
    for t, j in m.fs.time * m.fs.properties.component_list:
        if t <= 2:
            if j == "E":
                m.fs.cstr.inlet.conc_mol[t, j].fix(11.91 * 0.1 / 2.2)
            elif j == "S":
                m.fs.cstr.inlet.conc_mol[t, j].fix(12.92 * 2.1 / 2.2)
            elif j != "Solvent":
                m.fs.cstr.inlet.conc_mol[t, j].fix(0)
        elif t <= 4:
            if j == "E":
                m.fs.cstr.inlet.conc_mol[t, j].fix(5.95 * 0.1 / 2.2)
            elif j == "S":
                m.fs.cstr.inlet.conc_mol[t, j].fix(12.92 * 2.1 / 2.2)
            elif j != "Solvent":
                m.fs.cstr.inlet.conc_mol[t, j].fix(0)
        else:
            if j == "E":
                m.fs.cstr.inlet.conc_mol[t, j].fix(8.95 * 0.1 / 2.2)
            elif j == "S":
                m.fs.cstr.inlet.conc_mol[t, j].fix(16.75 * 2.1 / 2.2)
            elif j != "Solvent":
                m.fs.cstr.inlet.conc_mol[t, j].fix(0)

    m.fs.cstr.inlet.conc_mol[:, "Solvent"].fix(1.0)

    m.fs.cstr.inlet.flow_vol.fix(2.2)
    m.fs.cstr.inlet.temperature.fix(300)

    # Fix outlet conditions
    m.fs.cstr.outlet.flow_vol.fix(2.2)
    m.fs.cstr.outlet.temperature[m.fs.time.first()].fix(300)

    assert degrees_of_freedom(m) == 0

    initialize_by_time_element(m.fs, m.fs.time, solver=solver)

    assert degrees_of_freedom(m) == 0

    # Assert that the result looks how we expect
    assert m.fs.cstr.outlet.conc_mol[0, "S"].value == 0
    assert abs(m.fs.cstr.outlet.conc_mol[2, "S"].value - 11.389) < 1e-2
    assert abs(m.fs.cstr.outlet.conc_mol[4, "P"].value - 0.2191) < 1e-3
    assert abs(m.fs.cstr.outlet.conc_mol[6, "E"].value - 0.0327) < 1e-3
    assert abs(m.fs.cstr.outlet.temperature[6].value - 289.7) < 1

    # Assert that model is still fixed and deactivated as expected
    assert m.fs.cstr.control_volume.material_holdup[m.fs.time.first(), "aq", "S"].fixed

    for t in m.fs.time:
        if t != m.fs.time.first():
            assert not m.fs.cstr.control_volume.material_holdup[t, "aq", "S"].fixed

            assert not m.fs.cstr.outlet.temperature[t].fixed
        assert m.fs.cstr.control_volume.material_holdup_calculation[t, "aq", "C"].active

        assert m.fs.cstr.control_volume.properties_out[t].active
        assert not m.fs.cstr.outlet.conc_mol[t, "S"].fixed
        assert m.fs.cstr.inlet.conc_mol[t, "S"].fixed

    # Assert that constraints are feasible after initialization
    for con in m.fs.component_data_objects(Constraint, active=True):
        assert value(con.body) - value(con.upper) < 1e-5
        assert value(con.lower) - value(con.body) < 1e-5

    results = solver.solve(m.fs)
    assert check_optimal_termination(results)
